package facets.core.app;
import static facets.core.app.AppSurface.ContentStyle.*;
import static facets.util.app.Events.*;
import facets.core.app.AppSurface.ContentCreationException;
import facets.core.app.Headless.HeadlessHost;
import facets.core.superficial.FacetedTarget;
import facets.core.superficial.Notice;
import facets.core.superficial.SFacet;
import facets.core.superficial.SFrameTarget;
import facets.core.superficial.SIndexing;
import facets.core.superficial.STarget;
import facets.core.superficial.STargeter;
import facets.core.superficial.Notifying.Impact;
import facets.core.superficial.app.AreaTargeter;
import facets.core.superficial.app.SAreaTarget;
import facets.core.superficial.app.SContentRootTargeter;
import facets.core.superficial.app.SContenter;
import facets.core.superficial.app.SHost;
import facets.core.superficial.app.SSurface;
import facets.core.superficial.app.SHost.FacetLayout;
import facets.util.Debug;
import facets.util.Tracer;
import facets.util.Util;
import facets.util.app.WatchableOperation;
import facets.util.app.AppValues;
import facets.util.app.AppWatcher;
import facets.util.datatree.ValueNode;
import java.util.ArrayList;
import java.util.List;
/**
Implements {@link SSurface} for both single- and multi-content applications. 
<p>{@link AppSurface} maintains a set of {@link SContenter} members 
	and an internal surface area root containing content area roots exposing 
	these members; a wide range of methods enable addition and removal of 
	both roots and content. 
<p>Depending on {@link AppSpecifier#contentStyle()} in the {@link AppSpecifier} 
passed to the core constructor, {@link AppSurface} 
	can build an application whose content is exposed three ways:</p>
<ul>
	<li><code>SINGLE</code> - single content and content root
	<li><code>TABBED</code> - multiple content exposed by content 
		roots in folder tabs
	<li><code>DESKTOP</code> - multiple content (potentially of multiple types)
	roots exposed by windows within internal desktop
</ul>
<p>Once the surface is built, the <code>add/removeXXX</code> methods can 
	be called to perform operations on both the content roots and their underlying 
	{@link SContenter}s. 
 */
public abstract class AppSurface extends Tracer implements SSurface{
	/**
	Convenience constant denoting level in the {@link SAreaTarget} tree.  
	 */
	final public static int AREA_APP=0,AREA_CONTENT=1,AREA_PANE=2,AREA_PAGE=3;
	/**
	Title of {@link #emptyContent}. 
	 */
	public static final String TITLE_EMPTY="Empty Root";
	/**
	{@link SContenter} for use in an {@link AppSurface}. 
	<p>{@link Contenter} defines the extended contract that must be
	implemented by any {@link SContenter} supplying content for
	an {@link AppSurface}. 
	 */
	public interface Contenter extends facets.core.superficial.app.SContenter{
		/**
		Return a layout of facets that expose a content type. 
		@param host will generally create the layout, to ensure that it is of 
		a type known to {@link SHost#setLayout(FacetLayout)}
		 * @param appRoot is attached to the {@link SAreaTarget} heading 
		the application surface area tree
		 * @param t was created by a content root returned by 
		{@link SContenter#newContentRoot(boolean, boolean)}
		 */
		FacetLayout newContentLayout(SHost host,SFacet appRoot,SContentRootTargeter t);
		/**
		Allows the contenter to align two target trees, for instance when 
		'cloning' a window. 
		@param existing containing values to copy
		@param added new root to be aligned with existing
		 */
		void alignAddedRoot(SAreaTarget existing,SAreaTarget added);
		/**
		Has the content changed? 
		 */
		boolean hasChanged();
		/**
		Called when the content has been added to the surface.  
		 */
		void wasAdded();
		/**
		Called when the content has been removed from the surface.  
		 */
		void wasRemoved();
	}
	/**
	The possible styles for an {@link AppSurface}. 
	<p>Styles are defined in terms of content presentation: 
	<ul><li>{@link #SINGLE} - single (possibly replaceable) content
	<li>{@link #TABBED} - multiple content in tabs
	<li>{@link #DESKTOP} - multiple content in windows of internal desktop
	</ul>
	 */
	public static enum ContentStyle{
		SINGLE,TABBED,DESKTOP;
		public static final String KEY_NATURE="contentStyle";
		public String toString(){
			return name().toLowerCase();
		}
		public static ContentStyle getStyle(AppValues values){
			ValueNode nature=values.nature(true);
			String got=nature.getString(KEY_NATURE);
			if(got.equals(""))nature.put(KEY_NATURE,DESKTOP.toString());
			return got.equals(SINGLE.toString())?ContentStyle.SINGLE
					:got.equals(TABBED.toString())?ContentStyle.TABBED
							:ContentStyle.DESKTOP;
		}
	};
	/**
	Heads a content area tree for use in an application surface. 
	 */
	public static class ContentRoot extends SAreaTarget{
    private String useTitle=superTitle();
    private final SContenter contenter;
		public ContentRoot(String title,FacetedTarget[]viewers,SContenter contenter){
    	super(title,SIndexing.newDefault("Panes",viewers,0));
    	this.contenter=contenter;
    }
    public final STargeter newTargeter(){
      return new SContentRootTargeter(contenter);
    }
  	final void setUseTitle(String title){
  		useTitle=title;
  	}
  	final String superTitle(){
			return super.title();
		}
  	final public String title(){
  		if(contenter!=null){
				String contentTitle=contenter.title();
				if(!useTitle.startsWith(contentTitle))
					useTitle=useTitle.replaceAll("^[^:]+",contentTitle);
				if(false)Util.println("AppSurface.ContentRoot.title: "+contentTitle);
			}
  		return useTitle;
  	}
  }
  /**
  Non-existent content of an "empty" {@link ContentStyle#DESKTOP} 
  application. 
  <p>When returned by {@link AppSurface#findActiveContent()} serves as a flag.   
   */
  public final Contenter emptyContent=new Contenter(){
		public final SAreaTarget newContentRoot(boolean faceted, boolean shareSelection){
			return new SAreaTarget(title(),SAreaTarget.EMPTY_AREA_CHILDREN){
				public STargeter newTargeter(){
					return new SContentRootTargeter(this);
				}
			};
		}
		public FacetLayout newContentLayout(SHost host,
				SFacet appRoot,SContentRootTargeter t){
			return newEmptyDesktopLayout(host,t,appRoot);
		}
		public SFrameTarget contentFrame(){
			return new SFrameTarget("No contentFrame()");
		}
		public void alignAddedRoot(SAreaTarget existing,SAreaTarget added){
		}
		public boolean hasChanged(){
			return false;
		}
		public void rootRetargeted(SContentRootTargeter t){
		}
		final public String title(){
			return TITLE_EMPTY;
		}
		public void wasAdded(){}
		public void wasRemoved(){}
		public String toString(){return Debug.info(this);}
	};
	/** Immutable {@link ContentStyle} for the application.*/
	public final ContentStyle contentStyle;
	/** Immutable {@link AppSpecifier} for the application. */
	final public AppSpecifier spec;
	/** Immutable {@link AppValues} for the application. */
	final public AppValues values;
	private final AppRoots roots;
	private List<Contenter>contents=new ArrayList();
	/**
  Constructs an application surface. 
  <p>The surface may be empty if {@link #contentStyle} is <code>DESKTOP</code>
  	otherwise will be initialised with content constructed from return of 
  	{@link #getOpeningContentSources()}). 
	<p>The surface can form the basis of either a free-standing application that 
	  creates its own window and thus its own 
	  {@link facets.core.superficial.app.SHost.FacetLayout}, 
	  or one created within and by its {@link SHost} such as 
	  an applet. 
	<p>Application build sequence is as follows:</p>
	<ol>
	  <li>Create a suitable {@link AppSurface} in <code>main</code> 
	    if free-standing, in <code>init</code> if applet-hosted.</li>
	  <li>Call {@link #openApp()} on the {@link AppSurface}. </li>
	  <li>The {@link AppSurface} calls {@link SSurface#buildRetargeted()} to build facet 
	    layouts which it passes the host. </li>
	  </ol>
   */
  protected AppSurface(AppSpecifier spec){
  	if((this.spec=spec)==null)throw new IllegalArgumentException(
  			"Null spec in "+Debug.info(this));
  	else{
  		spec.attachToApp(this);
  		contentStyle=spec.contentStyle();
  		values=spec.values;
  	}
		traceEvent(">Created surface "+Debug.info(this));
  	roots=new AppRoots(this);
  }
	/**
  Create targets to be returned as the <code>elements</code> of the surface root. 
  @return a non-<code>null</code> (though possibly empty) {@link STarget}[] 
   */
  protected STarget[]lazySurfaceRootElements() {
		return new STarget[]{};
	}
	/**
  Defines an application and builds its surface. 
    <p>Calls the final implementation of {@link SSurface#buildRetargeted()}
   */
  public void openApp(){
  	traceEvent(">Opening app "+Debug.info(this));
  	buildRetargeted();
  	traceEvent(">Opened app "+Debug.info(this)+"\n");
  }
  /**
  Builds an appropriate surface for the type passed to the 
  constructor. 
  <p>Opening content is created from sources returned by {@link #getOpeningContentSources()}.
  <p>Surfaces have the following features:
  <ul>
    <li><code>SINGLE</code> - content root facet from 
      the single {@link SContenter} added directly to the 
      host GUI; menu, tool and status facet from the content 
      can be merged with roots defined by the {@link AppSurface}. 
    <li><code>TABBED</code> - one or more tabs containing 
      content root facets obtained from multiple {@link SContenter}s, 
      managed by the {@link AppSurface} which sets new 
      merged menu, tool and status roots as required for the 
      active tab. 
    <li><code>DESKTOP</code> - content root facet 
      in internal windows; <code>DESKTOP</code> as a singleton source
		signals that an empty desktop be created with suitable menus etc. 
  </ul>
   */
  final public void buildRetargeted(){
  	Object[]sources=getOpeningContentSources();
  	if(sources==null||sources.length==0)throw new IllegalStateException(
  			"Null or empty sources in "+Debug.info(this));
  	else if(contentStyle==DESKTOP&&sources[0]==DESKTOP)
  		contents.add(emptyContent);
		else for(int i=0;i<sources.length;i++)
			contents.add((Contenter)newContenter(sources[i]));
  	for(Contenter each:contents)
  		roots.new RootsUpdate().addFreshRoot(each);
  	if(isHeadless())roots.headless.doAppTest(this);
  	roots.updateLayout(contents);
  }
	public void notify(Notice notice){
		if(trace)traceEvent("Surface "+Debug.info(this)+" notified with "+notice);
	  Impact impact=notice.impact;
	  STargeter targeter=surfaceRootTargeter();
		if(!contents.isEmpty()&&!contents.contains(emptyContent)){
		  SContentRootTargeter rootThen=activeContentRootTargeter(),rootNow=null;
	    targeter.retarget(targeter.target(),impact);
	    rootNow=activeContentRootTargeter();
	    boolean updateLives=true;
			if(updateLives)rootNow.target().setLive(true);
			if(!rootNow.title().equals(TITLE_EMPTY))
				for(Contenter content:contents)
					if(content instanceof ViewerContenter
						&&((ViewerContenter)content).contentFrame()==rootNow.content().target())
						content.rootRetargeted(rootNow);
	    if(rootThen!=rootNow){
	    	if(updateLives)rootThen.target().setLive(false);
	  		roots.updateLayout(contents);
	    }
	  }
	  if(trace)traceEvent(("Surface tree retargeted in "+Debug.info(this)));
	  surfaceRetargeted();
	  targeter.retargetFacets(impact);
	  if(trace)traceEvent("Surface facets retargeted in "+Debug.info(this));
	}
	/**
	Return sources for content to be exposed when the surface is first created.
	<p>The objects returned must be suitable for passing to {@link #newContenter(Object)}.  
	@return suitable source[es] for opening content. 
	 */
	protected abstract Object[]getOpeningContentSources();
	/**
  Return a {@link SContenter} wrapping content created from 
  <code>source</code>. 
  <p>The content must be an implementor of {@link AppSurface.Contenter} such as 
  {@link ViewerContenter}; it will appear in the surface as follows: 
  	<ul>
  		<li><code>SINGLE</code> - replacing content in any existing surface</li>
  		<li><code>TABBED</code>, <code>DESKTOP</code> - in new window/tab</li>
  	</ul>
   @param source for the content
   */
  abstract protected SContenter newContenter(Object source);
  /**
  Provide a empty {@link FacetLayout}.
  <p>Implements {@link Contenter#newContentLayout(SHost, SFacet, SContentRootTargeter)}
  on behalf of {@link #emptyContent} 
   */
  protected abstract FacetLayout newEmptyDesktopLayout(SHost host,
  		SContentRootTargeter rootTargeter,SFacet surfaceRootFacet);
	/**
	May return a facet to be attached to the surface root. 
	<p>Defines a tab folder or internal desktop for a multi-content
	surface - not called for single content as the content root facet is used instead.
	@param surfaceRoot was (re)created in {@link SSurface#buildRetargeted()}
	 */
	protected MountFacet newMultiContentFacet(SAreaTarget surfaceRoot){
		throw new RuntimeException("Not implemented in "+Debug.info(this));
	}
	/**
	Opens or activates a window/tab on <code>source</code>. 
	<p>Where type is <code>TABBED</code> or <code>DESKTOP</code> 
		and <code>source</code> is different from that of any existing content, 
		creates a new {@link SContenter} and from that a new content root; 
		where <code>source</code> is already in use, 
		activates an existing content root exposing the {@link Contenter} 
		for that source. 
	<p>If client code throws a {@link ContentCreationException} the method returns without
	effect after passing the exception to {@link #contentNotAdded(ContentCreationException)}.  
	 @param source for added content
	 */
	final public void addContent(Object source){
	  Contenter content=(Contenter)newContenter(source);
	  try{
			roots.new RootsUpdate().addFreshRoot(content);
		}catch(ContentCreationException e){
			contentNotAdded(e);
			return;
		}
		contents.add(content);
		contents.remove(emptyContent);
		roots.updateLayout(contents);
		content.wasAdded();
	}
	/**
	Enables a client to abandon creation of content. 
	 */
	public static final class ContentCreationException extends RuntimeException{
		public ContentCreationException(String message){
			super(message);
		}
	}
  protected void contentNotAdded(ContentCreationException e){
		throw new RuntimeException(e);
	}
	/**
	Replaces content in single-content surface. 
	<p>Where type is <code>SINGLE</code>, replaces content in existing surface 
	with new {@link Contenter} created from <code>source</code>.
	  @param source for replacement content
	 */
	final public void replaceSingleContent(Object source){
		findActiveContent().wasRemoved();
	  Contenter content=(Contenter)newContenter(source);
	  contents=new ArrayList();
	  contents.add(content);
	  roots.new RootsUpdate().replaceActiveRoot(content);
	  if(isHeadless())return;
	  roots.updateLayout(contents);
  	content.wasAdded();
	}
	/**
	Restores content of active root. 
	<p>Any other roots for the same content are removed.  
	 */
	final public void revertActiveContent(){
	  Contenter contentThen=findActiveContent(),
	  	contentNow=(Contenter)newContenter(((ViewerContenter)contentThen).sink());
	  contents.remove(contentThen);
	  contentThen.wasRemoved();
	  contents.add(contentNow);
	  roots.new RootsUpdate().replaceActiveRoot(contentNow);
	  roots.updateLayout(contents);
	}
	/**
	Attempts to close all tabs/windows exposing content of the active viewer. 
	<p>Providing {@link #contentIsRemovable(Contenter)} returns <code>true</code> for 
		the content of the active viewer, closes all tabs/windows exposing that content.
	*/
	final public boolean removeActiveContent(){
	  SFrameTarget viewable=roots.surfaceRoot().activeContent();
	  List<Contenter>keepers=new ArrayList();
	  for(Contenter content:contents)
			if(((ViewerContenter)content).contentFrame()==viewable){
				if(!contentIsRemovable(content))return false;
				else{
					roots.new RootsUpdate().removeContentRoots(content);
					content.wasRemoved();
				}
			}
			else keepers.add(content);
	  contents=new ArrayList();
		contents.addAll(keepers);
		if(contents.isEmpty())contents.add(emptyContent);
		roots.updateLayout(contents);
		return true;
	}
	/**
	Attempts to close all tabs/windows. 
	<p>Removes content roots for all content for which 
	{@link #contentIsRemovable(Contenter)} returns <code>true</code>; 
	 */
	final public void removeAllContent(){
		if(contentStyle!=DESKTOP)throw new IllegalStateException(
				"Not " +DESKTOP.name()+" in "+Debug.info(this));
	  for(Contenter content:contents)
			if(contentIsRemovable(content))
	    	roots.new RootsUpdate().removeContentRoots(content);
		if(contents.isEmpty())contents.add(emptyContent);
		roots.updateLayout(contents);
	}
	/**
  Opens a new tab/window on existing content. 
  <p>Where type is <code>TABBED</code> or <code>DESKTOP</code>, 
  	adds a new content root exposing the content of the active content root.
   */
  final public void cloneActiveRoot(){
    int activeAt=roots.surfaceRoot().indexing().index();
    roots.new RootsUpdate().addAlignedRoot(findActiveContent(),activeAt);
    roots.updateLayout(contents);
  }
  /**
	Attempts to close tab/window containing active viewer. 
	<p>Where there is no other root for the content, fails if 
		{@link #removeActiveContent()} returns <code>false</code>.
	 */
	final public boolean removeActiveRoot(){
		SAreaTarget surfaceRoot=roots.surfaceRoot(),
			contentRoots[]=(SAreaTarget[])surfaceRoot.indexableTargets(),
			activeRoot=(SAreaTarget)surfaceRoot.indexedTarget();
	  SFrameTarget viewable=activeRoot.activeContent();
	  boolean isLastArea=true;
	  for(SAreaTarget root:contentRoots)
	  	if(root!=activeRoot)
	  		isLastArea&=root.activeContent()!=viewable;
		if(isLastArea)return removeActiveContent();
	  else{
	  	roots.new RootsUpdate().removeRoot(activeRoot);
	  	roots.updateLayout(contents);
	  }
		return true;
	}
  /**
  Returns the content that created the active {@link SAreaTarget} tree. 
  @return the current content or {@link #emptyContent}
   */
  final public Contenter findActiveContent(){
  	boolean debug=false;
	  SFrameTarget viewable=roots.surfaceRoot().activeContent();
		if(debug)traceDebug(".findActiveContent: ",viewable);
		for(Contenter content:contents){
			if(debug)traceDebug(".findActiveContent: ",content.contentFrame());
			if(content.contentFrame()==viewable)
	  		return content;
		}
		return emptyContent;
	}
	/**
	Returns the first {@link SAreaTarget} tree exposing the source passed. 
	@param source of content to be found (typically a file) 
	 */
	final public SAreaTarget firstContentRoot(Object source){
		Contenter found=null;
		for(Contenter content:contents)
			if(content instanceof ViewerContenter && 
	  			((ViewerContenter)content).sink().equals(source))
	  		found=content;
		if(found==null)return null;
		STarget[]roots=((SAreaTarget)surfaceRootTargeter().target()).indexableTargets();
		for(int i=0;i<roots.length;i++)
			if(((SAreaTarget)roots[i]).activeContent()
					==((ViewerContenter)found).contentFrame())
				return(SAreaTarget)roots[i];
		return null;
	}
	/**
	Check if  content passed may be removed. 
	<p>The default implementation checks {@link Contenter#hasChanged()}. 
	@param content to be checked
	 */
	public abstract boolean contentIsRemovable(Contenter content);
  /**
  Called whenever the surface root has been retargeted. 
   */
  protected void surfaceRetargeted(){}
  /**
	Returns the targeter of the currently active content root. 
	 */
	final public SContentRootTargeter activeContentRootTargeter(){
  	return(SContentRootTargeter)surfaceRootTargeter().area(AREA_CONTENT);
  }
	public final AreaTargeter surfaceRootTargeter(){
		if(roots.targeter==null)throw new IllegalStateException(
				"No targeter in "+Debug.info(this));
		return roots.targeter;
	}
	/**
	Implements interface method. 
	@return {@link AppValues#appName} from {@link #values}. 
	 */
	final public String title(){
		return values.appName;
	}
	/**
	Returns all {@link ViewerContenter}s currently active in the application. 
	 */
	public List<ViewerContenter>getViewerContents(){
		List<ViewerContenter>list=new ArrayList();
		for(Contenter each:contents)if(each instanceof ViewerContenter)
			list.add((ViewerContenter)each);
		return list;
	}
	/**
	Provides for an application to run headless. 
	@return <code>false</code> by default
	 */
	protected boolean isHeadless(){
		return false;
	}
	/**
	Provides the host for a headless application. 
	@return a {@link Headless.HeadlessHost}
	 */
	final protected SHost newHeadlessHost(){
		return new HeadlessHost(this);
	}
	/**
	Allows a check that the surface has been built. 
	@return <code>true</code> once {@link #surfaceRootTargeter()} does not
	throw an {@link IllegalStateException}
	 */
	final public boolean isBuilt(){
		return roots.targeter!=null;
	}
	/**
	Allows runtime exceptions in framework and client code to be handled gracefully, 
	by {@link AppWatcher}.
	<p>Default simply runs <code>r</code> 
	 */
	public void runWatched(WatchableOperation r){
		r.doSimpleOperation();
	}
}/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
