package facets.core.app;
import static facets.core.app.AppConstants.*;
import static facets.core.app.DialogPages.*;
import static facets.core.app.Dialogs.*;
import static facets.util.Debug.*;
import static facets.util.app.AppValues.*;
import static facets.util.app.Events.*;
import facets.core.app.DialogSurface.Contenter;
import facets.core.superficial.FacetedTarget;
import facets.core.superficial.Notice;
import facets.core.superficial.Notifiable;
import facets.core.superficial.Notifying;
import facets.core.superficial.SFacet;
import facets.core.superficial.SFrameTarget;
import facets.core.superficial.SIndexing;
import facets.core.superficial.STarget;
import facets.core.superficial.STargeter;
import facets.core.superficial.STrigger;
import facets.core.superficial.TargetCore;
import facets.core.superficial.Notifying.Impact;
import facets.core.superficial.app.AreaTargeter;
import facets.core.superficial.app.IndexingTarget;
import facets.core.superficial.app.SAreaTarget;
import facets.core.superficial.app.SContentRootTargeter;
import facets.core.superficial.app.SContenter;
import facets.core.superficial.app.SSurface;
import facets.core.superficial.app.ViewableFrame;
import facets.core.superficial.app.SHost.FacetLayout;
import facets.util.Debug;
import facets.util.Times;
import facets.util.Titled;
import facets.util.Tracer;
import facets.util.Util;
import facets.util.app.AppValues;
import facets.util.app.Events;
import facets.util.datatree.TypedNode;
import facets.util.datatree.ValueNode;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
/**
Implements {@link SSurface} for modal dialogs. 
<p>{@link DialogSurface} can build different dialog styles 
  depending on the {@link Contenter}s passed to its constructor 
  and the methods overriden in the instance subclass.
<ul>
  <li><b>simple </b> - pass a single {@link Contenter}, 
    from which a content root is created and added as the single 
    page. 
  <li><b>tabbed </b> - pass multiple {@link Contenter}s, 
    from which the content roots are added as children of a single 
    page. 
  <li><b>multi-page</b> - create tree members whose {@link SAreaTarget}
    values are added as pages. 
</ul>
 */
public abstract class DialogSurface extends DialogPages implements SSurface{
	public final static int TRIMMINGS_TAB=20,TRIMMINGS_TREE=100;
	public static final int AT_NOT_SET=Integer.MIN_VALUE;
	/**
	Heads a content area tree for use in a dialog page. 
	 */
  public static class ContentRoot extends SAreaTarget{
    protected static final boolean debug=false;
		private final SContenter contenter;
		public ContentRoot(String title,STarget[]frames,SContenter contenter){
    	super(title,frames);
    	this.contenter=contenter;
    }
    public final STargeter newTargeter(){
      return new SContentRootTargeter(contenter){
      	private SIndexing storeAreasIndexing;
        private boolean builtFacets=false;
        public void retarget(STarget target,Notifying.Impact impact){
        	super.retarget(target, impact);
		  		SAreaTarget root=(SAreaTarget)target;
        	if(builtFacets||!(root.indexedTarget()instanceof FacetedTarget))return;
      		storeAreasIndexing=root.indexing();
      		SIndexing preBuild=SIndexing.newDefault(root.title(),
      				new SFrameTarget[]{new SFrameTarget("Pre-build",
      						contenter.contentFrame().framed)},0);
      		root.setIndexing(preBuild);
      		if(debug)finishBuild(root);
        }
				public void retargetFacets(Notifying.Impact impact){
					super.retargetFacets(impact);
		  		SAreaTarget root=(SAreaTarget)target();
					if(!builtFacets&&root.isLive())finishBuild(root);
				}
				private void finishBuild(SAreaTarget root){
					if(storeAreasIndexing!=null)root.setIndexing(storeAreasIndexing);
		  		SFacet contentPanel=((Contenter)contenter).newContentPanel(this);
		  		if(contentPanel==null)throw new RuntimeException(
		  				"Illegal null content in "+Debug.info(this));
		  		((MountFacet)root.attachedFacet()).setFacets(new SFacet[]{contentPanel});
		  		builtFacets=true;
				}
			};
    }
  }
	/**
	{@link SContenter} for use in a {@link DialogSurface}. 
	<p>{@link Contenter} defines an extended contract that must be
	implemented by any {@link SContenter} supplying content for
	a {@link DialogSurface}. 
	 */
	public interface Contenter extends SContenter{
		/**
		Return a panel containing facets exposing the content. 
		<p>Called from {@link STargeter#retargetFacets(Notifying.Impact)} in the (special)
		{@link SContentRootTargeter} created by {@link DialogSurface.ContentRoot}
		@param t heads the targeter tree for the content
		 */
		SFacet newContentPanel(SContentRootTargeter t);
		/**
		Enables the contenter to interact with its host dialog. 
		<p>Can be implemented empty where no interaction is required. 
		 */
		void setDialog(DialogSurface dialog);
		/**
		Return the panel size needed for the content surface elements. 
		 */
		Dimension contentAreaSize();
		/**
		Return a title for the containing page. 
		 */
		String title();
		/**
		Change content state to match surface state. 
		<p>This will generally mean setting content state to that of a buffer
		storing surface state. 
		 */
		void applyChanges();
		/**
		Restore content to its original state. 
		 */
		void reverseChanges();
		/**
		Called once the containing dialog has closed. 
		 */
		void dialogClosed();
	}
	/**
	Defines top-level buttons for a {@link DialogSurface}. 
	 */
	public static abstract class DialogActions{
		private DialogSurface surface;
		/**
		Called from {@link DialogSurface#DialogSurface(ActionAppSurface, DialogHost, String, 
			DialogSurface.DialogActions, DialogSurface.Contenter[])}. 
		 */
		protected void attachSurface(DialogSurface surface){
			this.surface=surface;
		}
		/**
		Allows subclasses to access the {@link DialogSurface}. 
		@return the surface attached with {@link #attachSurface(DialogSurface)}
		 */
		protected final DialogSurface surface(){
			if(surface==null)throw new RuntimeException("Null surface in "+Debug.info(this));
			else return surface;
		}
		/**
		Create {@link STrigger}s defining the dialog control buttons. 
		 */
		public abstract STrigger[]newTriggers();
	}	
	/**
	Creates a key representing the exact types of the {@link SContenter}s
	passed. 
	<p>Used to implement {@link #replaceContents(Contenter[])}; 
	available for use by
	any class requiring a definition of the composite type of a {@link SContenter}[]
	@param contents define the key returned by their types
	 */
	public final static Object contentsKey(Contenter[]contents){
		StringBuffer b=new StringBuffer();
		for(int i=0;i<contents.length;i++)b.append(contents[i].getClass().getName());
		return b.toString();
	}
	public final ActionAppSurface app;
	private final ValueNode localState=new ValueNode(AppConstants.STATE_NODE_DIALOGS,"localState");
	private final DialogActions actions;
	private final String keyTail=Util.shortTypeNameKey(this),boundsKey=KEYTOP_BOUNDS+keyTail,
		sizeDefaultKey=KEYTOP_NATURE_SIZE+keyTail;
  private Contenter[]contents;
	private Rectangle boundsState,boundsDefault;
	/**
	Unique constructor. 
	@param app exposed in this surface; stored as {@link #app}
	@param host the {@link DialogHost} managing the dialog window 
	@param title passed to the host dialog window
	@param actions the {@link DialogActions} defining a set of control actions for the dialog
	@param contents the {@link Contenter}s defining the content to be exposed in the dialog 
	 */
	public DialogSurface(ActionAppSurface app,DialogHost host,String title,
			DialogActions actions,Contenter[]contents){
		super(host,title);
		this.contents=contents;
		if((this.actions=actions)==null)throw new IllegalArgumentException(
				"Null actions in "+Debug.info(this));
		else if((this.app=app)==null)throw new IllegalArgumentException(
				"Null app in "+Debug.info(this));
		actions.attachSurface(this);
		host.setTitle(title);
	}
	/**
	Builds the dialog surface. 
	<p> All surface variants are built by the same code, with execution path 
	determined primarily by whether it builds a single
	or multiple page layout. </p>
	<ol>
		<li>Content roots are constructed and their facet trees built for each of 
		the {@link Contenter}s passed to the constructor, 
		and passed to {@link #newContentPages(DialogSurface.Contenter[], SAreaTarget[])}.</li>
		<li> Facets are attached using {@link #attachPageFacet(SAreaTarget)} to 
		any pages returned that are not content roots 
		(and thus already have facet attached). </li>
		  <li>A single page is contained directly by the root area 
	      passed to {@link #attachSurfaceRootFacets(SAreaTarget)}.</li>
		  <li>Multiple pages are passed to {@link #newPageTreeNodes(SAreaTarget[])} 
	      which defines nodes for a page selection tree, and these 
	      nodes passed to {@link #newPagesTreeView(TypedNode[])}; 
	      from this a viewer area is created which is one of two 
	      children of the root area, the other being an area containing 
	      the pages.</li>
	</ol>
	 */
	final public void buildRetargeted(){
		rebuild(contents);
	}
	/**
	Replaces all content in the surface. 
	<p>Existing content roots are replaced by roots created by 
		each member of <code>contents</code>, whose types must match 
		exactly those of the existing contents. 
		@param contents replace existing contents 
	 */
	public void replaceContents(Contenter[]contents){
		if(this.contents==null||contents==null)
			throw new IllegalArgumentException("Null contents in "+Debug.info(this));
		if(!contentsMatch(this.contents,contents))
			throw new IllegalArgumentException("Contents do not match existing in "
					+Debug.info(this));
		if(false)for(Contenter c:this.contents)
			traceDebug("DialogSurface.replaceContents: ",c.contentFrame());
		rebuild(this.contents=contents);
	}
	final void rebuild(Contenter[]contents){
		for(Contenter contenter:contents)contenter.setDialog(this);
		super.rebuild(contents);
	}
	/**
	Implements interface method. 
	 */
	public void notify(Notice notice){
		if(trace)traceEvent("Surface "+Debug.info(this)+" notified with "+notice);
		Notifying.Impact impact=notice.impact;
		Notifying latest=notice.sources.get(notice.sources.size()-1);
		if(true&&((Titled)latest).title().equals(TYPE_PAGES_TREE))return;
		AreaTargeter targeter=surfaceRootTargeter();
		targeter.retarget(targeter.target(),impact);
		if(trace)traceEvent("Surface root "+Debug.info(targeter)+" retargeted");
	  AreaTargeter moveUp=targeter.area(SAreaTarget.AREA_LOWEST);
	  if(moveUp.title().equals(TYPE_PAGES_TREE))
	  	moveUp=(AreaTargeter)((SAreaTarget)((TypedNode)
	  			((NodeViewable)((SAreaTarget)moveUp.target()).activeContent()
	  					).selection().selection()).values()[0]).monitor();
	  while(!(moveUp instanceof SContentRootTargeter))
			moveUp=(AreaTargeter)moveUp.monitor();
	  SContentRootTargeter contentTargeter=(SContentRootTargeter)moveUp;
		activeContentType=contentTargeter.targetType;
		if(false)trace(".notify: ",activeContentType);
	  SAreaTarget activeRoot=(SAreaTarget)moveUp.target();
		activeRoot.setLive(true);
		for(Contenter c:contents){
			if(c.getClass()==activeContentType){
				if(false)trace(".notify: "+"c="+c.title());
				c.rootRetargeted(contentTargeter);
				if(false)host().setTitle(c.title());
			}
		}
	  if(trace)traceEvent(("Surface tree retargeted in "+Debug.info(this)));
	  targeter.retargetFacets(impact);
	  if(trace)traceEvent(("Surface facets retargeted in "+Debug.info(this)));
	}
	/**
	Implements abstract/interface method. 
	 */
	final STarget lazyTriggerGroup(){
		STarget[]triggers=actions.newTriggers();
		if(triggers==null)throw new IllegalStateException("Null triggers in "+Debug.info(this));
		return new TargetCore("Actions",triggers);
	}
	private ValueNode dialogValues(boolean nature){
		return true?nature?app.values.nature(true)
				:app.values.state(PATH_CORE,STATE_NODE_DIALOGS)
			:localState;
	}
	/**
	Allows the host to calculate/retrieve its bounds before launch.
	<p>Default size may be retrieved from {@link AppValues#nature(boolean)} using
	keys headed with {@link Dialogs#KEYTOP_NATURE_SIZE}; and 
	session bounds from {@link AppValues#state(String...)} using
	{@link Dialogs#KEYTOP_BOUNDS}.  
	@param areaRoot root of the {@link SAreaTarget} content tree 
	 */
	final public Rectangle getLaunchBounds(SAreaTarget areaRoot){
		if(boundsDefault==null){
			int width=0,height=0;
			for(Contenter content:contents){
				Dimension size=content.contentAreaSize();
				trace(".getLaunchBounds: content="+content.title()+" size="+size.height);
				width=Math.max(width,size.width);
				height=Math.max(height,size.height);
			}
			boolean multiPage=isMultiPage();
			width+=!multiPage?0:TRIMMINGS_TREE;
			STarget[]rootChildren=areaRoot.indexableTargets(),
				panelRoots=!multiPage?rootChildren:
					((IndexingTarget)rootChildren[MULTI_AREA_PAGES]).indexableTargets();
			for(STarget panelRoot:panelRoots)
				if(!(panelRoot instanceof SAreaTarget))height+=TRIMMINGS_TAB;
			int[]natureSize=dialogValues(true).getInts(sizeDefaultKey);
			Dimension sizeDefault=natureSize.length<2?new Dimension(width,height)
				:new Dimension(natureSize[0],natureSize[1]);
			boundsDefault=new Rectangle(new Point(AT_NOT_SET,0),sizeDefault);
			trace(".getLaunchBounds: natureSize="+natureSize.length+
					" boundsDefault="+boundsDefault.height);
		}
		if(boundsState==null){
			int[]ints=dialogValues(false).getInts(boundsKey);
			if(ints.length>3)boundsState=new Rectangle(ints[0],ints[1],ints[2],ints[3]);
			trace(".getLaunchBounds: boundsKey="+boundsKey+
					" boundsState="+(boundsState==null?null:boundsState.height));
		}
		Rectangle launch=boundsState=boundsState==null?boundsDefault
			:boundsState.width<boundsDefault.width
				||boundsState.height<boundsDefault.height?
			new Rectangle(boundsState.getLocation(),boundsDefault.getSize())
			:boundsState;
		trace(".getLaunchBounds: launch=",launch.height);
		return launch;
	}
	/**
	Allows the surface tidy up upon dismissal of its host.
	@param bounds latest of host window  
	 */
	final public void hostWindowDismissed(Rectangle bounds){
		for(int i=0;i<contents.length;i++)contents[i].dialogClosed();
		trace(".hostWindowDismissed: bounds="+bounds.getSize().height);
		if(bounds.x!=AT_NOT_SET)boundsState=new Rectangle(bounds);
		if(!persistBoundsState())return;
		trace(".hostWindowDismissed: boundsState="+boundsState.height);
		int[]ints={bounds.x,bounds.y,bounds.width,bounds.height};
		dialogValues(false).put(boundsKey,ints);
	}
	/**
	Should host bounds be persisted? 
	@return <code>true</code> by default
	 */
	public boolean persistBoundsState(){return true;}
	/**
	Enables top-level buttons to dismiss the dialog host. 
	 */
	public final void dismissHostWindow(){((DialogHost)host()).dismissWindow();}
	/**
	Required to ensure correct calculation of default host size. 
	<p>Default returns <code>false</code>.
	 */
	protected boolean isMultiPage(){return false;}
	/**
	Do the types of the two arrays match? 
	<p>Called by {@link #replaceContents(Contenter[])}; 
	implemented with {@link #contentsKey(Contenter[])}. 
	@param existing set during initial build or last launch
	@param proposed for this launch
	 */
	final protected boolean contentsMatch(Contenter[]existing,
			Contenter[]proposed){
		return contentsKey(existing).equals(contentsKey(proposed));
	}
	/**
	The contenters for the surface. 
	@return contenters passed during construction or to <code>replaceContents</code>
	 */
	public final Contenter[]contents(){
		if(contents==null)throw new IllegalStateException("No contents in "+Debug.info(this));
		return contents;
	}
	public static STarget findDialogTrigger(SContentRootTargeter rootTargeter,String title){
		for(STarget t:findRootArea(rootTargeter).elements()[0].elements())
			if(t.title().equals(title))return t;
		return STarget.NONE;
	}
	/**
	Convenience method for accessing the surface area root. 
	@param rootTargeter within the area targeter tree
	 */
	public static final SAreaTarget findRootArea(SContentRootTargeter rootTargeter){
		STargeter maybeRoot=(STargeter)rootTargeter.monitor();
		Notifiable check=maybeRoot;
		while((check=maybeRoot.monitor())instanceof STargeter)
			maybeRoot=(STargeter)check;
		return(SAreaTarget)maybeRoot.target();
	}
}
/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

