package facets.core.app;
import facets.core.app.DialogSurface.Contenter;
import facets.core.app.DialogSurface.DialogActions;
import facets.core.superficial.app.SContenter;
import facets.core.superficial.app.SSurface;
import facets.util.FileSpecifier;
import facets.util.Tracer;
import facets.util.app.WatchableOperation;
import java.io.File;
import java.io.FileFilter;
import java.util.HashMap;
import java.util.Map;
/**
Builds and launches modal dialogs on behalf of an {@link AppSurface}. 
<p>{@link Dialogs} launches dialogs of two kinds:</p>
<ul>
  <li>general-purpose input dialogs
  <li>dialogs hosting client-defined {@link DialogSurface}s as described for 
     {@link #launchSurfaced(Surfacer, String, DialogSurface.DialogActions, DialogSurface.Contenter...)}
</ul>
<p>{@link Dialogs} instances will be created by subclasses of {@link AppSurface}.  
 */
public abstract class Dialogs extends Tracer{
	/**
	Type-safe definitions of possible dialog responses. 
	 */
	public static enum Response{Ok,Cancel,Yes,No}
	final public static String DEBUG_RESIZABLE="dialogsAllResizable",
		KEYTOP_NATURE_SIZE="dialogSize",KEYTOP_BOUNDS="bounds";
	/**
	Creates the {@link DialogSurface} to expose a set of {@link SContenter}s.
	  */
	public static abstract class Surfacer{
		/**
		Create a {@link DialogSurface} to expose any {@link Contenter}s type-identical to
		those passed. 
		@param app application surface
		@param host dialog window for the surface to be created
		@param title for the dialog window
		@param actions defining top-level dialog buttons
		@param contents the initial contenters; can be replaced in
		the surface by any type-identical array
		 */
		protected abstract DialogSurface newSurface(SSurface app,
				DialogHost host,String title,DialogActions actions,Contenter[]contents);
	}
	public final AppSurface app;
	private final Map<Object,DialogSurface>surfaces=new HashMap();
	private boolean dialogOpen;
	protected Dialogs(AppSurface app){
		this.app=app;
	}
	/**
	 Launches a modal dialog, building its surface where required. 
	 <p>The surface built by the {@link Surfacer} on the first invocation 
	 for the {@link Contenter}[] 
	 array is stored for use in subsequent invocations with equivalent arrays. 
	  @param surfacer constructs the surface on first invocation
	  @param title passed to the surfacer
	  @param actions passed to the surfacer
	  @param contents passed either to the surfacer or to
	 {@link DialogSurface#replaceContents(DialogSurface.Contenter[])} in a previously-built
	 surface
	 */
	public final void launchSurfaced(final Surfacer surfacer,final String title, 
			final DialogActions actions,final DialogSurface.Contenter...contents){
		WatchableOperation op=new WatchableOperation("Dialogs.launchSurfaced"){
		@Override
		public void doSimpleOperation(){
			Object surfaceKey=DialogSurface.contentsKey(contents)
				+surfacer.getClass().getName();
			DialogSurface surface=surfaces.get(surfaceKey);
			if(surface==null){
				surface=surfacer.newSurface(app,newHost(),title,actions,
						contents);
				surface.buildRetargeted();
				surfaces.put(surfaceKey,surface);
			}
			else surface.replaceContents(contents);
			dialogOpen=true;
			((DialogHost)surface.host()).launchWindowedSurface(surface,app);
			dialogOpen=false;
		}};
		if(false)app.runWatched(op);
		else op.doSimpleOperation();
	}
	/**
	Is a dialog surface already open? 
	<p>Only applies to dialogs launched with
     {@link #launchSurfaced(Surfacer, String, DialogSurface.DialogActions, DialogSurface.Contenter...)}
	 */
	public boolean dialogOpen(){
		return dialogOpen;
	}
	/**
	Enables concrete subclasses to create an appropriate {@link DialogHost}
	for use by 
     {@link #launchSurfaced(Surfacer, String, DialogSurface.DialogActions, DialogSurface.Contenter...)}
	 */
	protected abstract DialogHost newHost();
	public abstract Response confirmOKCancel(String title,String rubric);
	public abstract Response confirmYesNo(String title,String rubric);
	public abstract Response warningYesNo(String title,String rubric);
	public abstract Response warningYesNoCancel(String title,String rubric);
	/**
	Parameter object for {@link Dialogs#warningException(ExceptionTexts, Exception, boolean)}. 
	 */
	public static final class ExceptionTexts{
		final public String title,rubricTop,rubricTail,detailTop;
		public ExceptionTexts(String title,String rubricTop,String rubricTail,
				String detailTop){
			this.title=title;
			this.rubricTop=rubricTop;
			this.rubricTail=rubricTail;
			this.detailTop=detailTop;
		}
		public ExceptionTexts(String appTitle,Exception e){
			this(appTitle+" - Exception","An exception has occurred: ",e.getMessage(),"");
		}
	}
	public abstract Response warningException(ExceptionTexts tt,Exception e,boolean inOpen);
	public abstract void infoMessage(String title,String rubric);
	public abstract void errorMessage(String title,String rubric);
	public abstract String getTextInput(String title,String rubric,String proposal);
	public abstract File openFile(FileSpecifier[]filters);
	public abstract File saveFile(File proposed,FileSpecifier[]filters);
	public abstract File confirmGetFile(String title, String path);
	public abstract void textPane(String title,String rubric, int rows, int cols);
	public abstract void htmlPane(String title,String path);
	final public int deleteDirFiles(final String title,final File dir,
			final FileSpecifier specifier){
		File[]files=dir.listFiles(new FileFilter(){
			@Override
			public boolean accept(File file){
				return specifier.specifies(file);
			}
		});
		String rubric=specifier.rubric;
		if(files.length==0){
			infoMessage(title,"No "+rubric +" found in '"+dir+"'.");
			return 0;
		}
		if(warningYesNo(title,"Are you sure you want to delete all "+rubric+
				" in '"+dir+"'?")!=Response.Yes)return -1;
		int deleteds=0;
		for(File file:files){
			file.delete();
			deleteds++;
		}
		infoMessage(title,"Deleted all " +deleteds+" "+rubric+" in '"+dir+"'.");
		return deleteds;
	}
}
/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

