package facets.facet;

import facets.util.Debug;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import javax.swing.JComponent;
import javax.swing.Timer;

/**
	Requests focus for a Swing component. 
	<p>{@link SwingFocusRequester} encapsulates the following process:
	<ol><li>requesting focus for a component
	<li>checking to see if focus has in fact been acquired
	<li>repeating the cycle as necessary
	<li>giving up if too many requests are made
</ol>
	 */
	class SwingFocusRequester{
		private int maxRequests;
		private int requests;
		private final JComponent swing;
		private final Timer requester=new Timer(requestInterval(),
				new ActionListener(){
			public void actionPerformed(ActionEvent e){
				if(++requests>maxRequests){
					if(false)throw new RuntimeException(
							"Too many focus requests for "+Debug.info(swing));
					else requester.stop();
				}
				swing.requestFocusInWindow();
			}
		});
		private final FocusAdapter listener=new FocusAdapter() {
			public void focusGained(FocusEvent e){
				requester.stop();
			}
		};
		/**
		Unique constructor. 
		@param swing to receive the focus
		 */
		public SwingFocusRequester(JComponent swing){
			(this.swing=swing).addFocusListener(listener);
		}
		/**
		Starts a request cycle.
		<p>Returns immediately if {@link JComponent#isFocusOwner()} is <code>true</code>;
		otherwise calls {@link JComponent#requestFocusInWindow()} immediately, and 
		thereafter as specified by {@link #requestInterval()} until its internal
		{@link FocusListener} receives a {@link FocusEvent} from the component
		passed to the constructor.    
		 */
		final public void startRequesting(){
			if(swing.isFocusOwner())return;
			swing.requestFocusInWindow();
			maxRequests=maxRequests();
			requests=1;
			requester.start();			
		}
		/**
		The maximum number of focus requests to be made. 
		<p>Requesting will stop if this number is exceeded. 
		 */
		protected int maxRequests(){
			return 5;
		}
		/**
		The interval to be used between focus requests.
		 */
		protected int requestInterval(){
			return 300;
		}
	}/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
