package facets.facet.app;
import static facets.core.app.AppConstants.*;
import facets.core.app.AppHost;
import facets.core.app.AppSurface;
import facets.core.superficial.app.SHost;
import facets.util.Debug;
import facets.util.Tracer;
import facets.util.datatree.ValueNode;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
/**
{@link AppHost} that manages a window for a {@link FacetAppSurface}. 
<p>{@link WindowAppHost} extends its superclass to manage 
<ul>
<li>opening and closing of the application and its host window
<li>validation and persistance of the window bounds
<li>an opening splash screen
</ul> 
<p>Concrete subclasses will define an appropriate event loop for their GUI toolkit. 
*/
public abstract class WindowAppHost extends Tracer implements AppHost{
	protected final FacetAppSurface app;
	protected WindowAppHost(FacetAppSurface app){
		this.app=app;
	}
	/**
	Store the current bounds (if not maximised and at least minimum size). 
	 */
	final protected void storeCheckedBounds(){
		Rectangle bounds=windowBounds();
		Dimension size=bounds.getSize();
		if(size.width>=screenSize().width)return;
	  Dimension min=minWindowSize();
		if(size.height<min.height||size.width<min.width)
			bounds=new Rectangle(bounds.getLocation(),minWindowSize());
		getBoundsState().put(STATE_ODDS_BOUNDS,new int[]{bounds.x,bounds.y,bounds.width,
				bounds.height});
	}
	/**
	Set the window bounds programmatically. 
	<p>Called from {@link #openWindow()}. 
	 */
	abstract protected void setWindowBounds(Rectangle bounds);
	/**
	Supply the current window bounds. 
	<p>Called from {@link #storeCheckedBounds()} and by 
	{@link FacetDialogSurface#calculateSmartDialogAt(Dimension, SHost)} 
	 */
	protected abstract Rectangle windowBounds();
	/**
	Supply the current screen size.
	<p>Called from {@link #storeCheckedBounds()} and by 
	{@link FacetDialogSurface#calculateSmartDialogAt(Dimension, SHost)} 
	 */
	protected abstract Dimension screenSize();
	/**
	Open the application window. 
	<p>Called from {@link AppSurface#buildRetargeted()} once the abstract surface
	is ready for conversion by the toolkit to a GUI. 
	<p>Default implementation just sets checked window bounds, concrete classes should
	override as required to create and open the application window.  
	 */
	protected void openWindow(){
		int[]ints=getBoundsState().getInts(STATE_ODDS_BOUNDS);
		if(ints.length==0){
			ValueNode natureCore=app.values.nature(true);
			ints=natureCore.getInts(NATURE_SIZE_DEFAULT);
			if(ints.length==0)ints=natureCore.getInts(NATURE_SIZE_MIN);
			if(ints.length<2)throw new IllegalStateException(
					"No sizes in "+Debug.info(natureCore));
			ints=new int[]{0,0,ints[0],ints[1]};
		}
		Dimension min=minWindowSize();
		final int X=0,Y=1,WIDTH=2,HEIGHT=3,BOUNDS=4;
		Rectangle check=new Rectangle(ints[X],ints[Y],ints[WIDTH],ints[HEIGHT]),
			bounds=check.width>=min.width&&check.height>=min.height?
		check:new Rectangle(new Point(check.x,check.y),min);
		setWindowBounds(bounds);
	}
	private ValueNode getBoundsState(){
		return app.values.state(FacetAppValues.PATH_CORE_APP);
	}
	/**
	Allow the application to close itself. 
	<p>Should simulate pressing the Close button. 
	 */
	protected abstract void requestAppClose();
	private Dimension minWindowSize(){
		int[]ints=app.values.nature(true).getInts(NATURE_SIZE_MIN);
	  final int WIDTH=0,HEIGHT=1,SIZE=2;
	  return new Dimension(ints[WIDTH],ints[HEIGHT]);
	}
	/**
	Put up a splash screen while the application is loading. 
	@param msg should be displayed on the screen
	 */
	protected abstract void splashUp(String msg);
	/**
	Take down any splash screen. 
	 */
	protected abstract void splashDown();
}/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
