package facets.facet.app.tree;
import static facets.facet.app.tree.TreeTargets.*;
import facets.core.app.AppHost.LayoutFeatures;
import facets.core.app.AppActions;
import facets.core.app.ListView;
import facets.core.app.MenuFacets;
import facets.core.app.NodeViewable;
import facets.core.app.TableView;
import facets.core.app.TreeView;
import facets.core.app.ViewerContenter;
import facets.core.superficial.SFacet;
import facets.core.superficial.SFrameTarget;
import facets.core.superficial.SIndexing;
import facets.core.superficial.STarget;
import facets.core.superficial.STargeter;
import facets.core.superficial.app.SAreaTarget;
import facets.core.superficial.app.SContentRootTargeter;
import facets.core.superficial.app.SContenter;
import facets.core.superficial.app.SHost;
import facets.core.superficial.app.SView;
import facets.core.superficial.app.ViewableFrame;
import facets.facet.FacetFactory;
import facets.facet.app.FacetAppSpecifier;
import facets.facet.app.FacetAppSurface;
import facets.facet.app.FacetAppValues;
import facets.idiom.XmlView;
import facets.util.Debug;
import facets.util.FileSpecifier;
import facets.util.datatree.DataConstants;
import facets.util.datatree.DataNode;
import facets.util.datatree.Nodes;
import facets.util.datatree.TypedNode;
import facets.util.datatree.XmlPolicy;
import facets.util.datatree.XmlSpecifier;
/**
{@link FacetAppSpecifier} for simple applications with {@link DataNode} content. 
<p>{@link TreeAppSpecifier} exemplifies practical use of the superclass with
a {@link ViewerContenter} (here {@link TreeAppContenter}) that provides the detail functionality.  
It also demonstrates with {@link #newContentRootTargets()} and
{@link #newTreeMenuItems(FacetFactory, STargeter[], STargeter[])} the key  
<a href="http://superficial.sourceforge.net/">Superficial</a>  mechanism for binding 
data abstracted as {@link STarget}s to widgets abstracted as {@link SFacet}s 
via an automatically-generated {@link STargeter} tree.  
<p>Concrete applications can be readily created by extending {@link TreeAppSpecifier} 
and types such as: 
<ul>
<li>{@link XmlPolicy} (especially {@link XmlPolicy#getTitleAttributeNames()})
and {@link XmlSpecifier}
<li>{@link SView}s such as {@link TreeView}, {@link ListView}, {@link TableView} 
for appropriate content 
</ul>
@see XmlView
 */
public abstract class TreeAppSpecifier extends FacetAppSpecifier{
	/**
	Key for integer defining size and style of 
	the return of {@link #getInternalContentSource()} 
	 */
	private static final String ARG_TREE_SIZE="treeSize";
	/**
	Unique constructor. 
	@param values passed to superclass
	 */
	public TreeAppSpecifier(FacetAppValues values){
		super(values);
	}
	/**
	Locks superclass implementation. 
	 */
	@Override
	final public AppActions newActions(){
		return super.newActions();
	}
	/**
	Final implementation. 
	<p>The {@link FacetAppSurface} returned calls out to {@link #getInternalContentSource()} 
	and to {@link XmlPolicy#fileSpecifiers()} via {@link #xmlPolicy()}; it creates
	 {@link TreeAppContenter}s. 
		 */
	@Override
	final protected FacetAppSurface newApp(FacetFactory ff){
		final TreeAppSpecifier thisSpec=this;
		return new FacetAppSurface(this,ff){
			@Override
			public FileSpecifier[]getFileSpecifiers(){
				return thisSpec.xmlPolicy().fileSpecifiers();
			}
			@Override
			protected SContenter newContenter(Object source){
				return new TreeAppContenter(source,(FacetAppSurface)app());
			}
			@Override
			protected Object getInternalContentSource(){
				return thisSpec.getInternalContentSource();
			}
		};
	}
	/**
	Defines an {@link XmlPolicy} for the application content. 
	<p>Called by the return of {@link #newApp(FacetFactory)} 
	and by {@link TreeAppContenter#saveToSink(Object)}.  
	@return by default a plain {@link XmlPolicy}
	 */
	protected XmlPolicy xmlPolicy(){
		return new XmlPolicy(){
			@Override
			protected boolean treeAsXmlRoot(){
				return false?true:super.treeAsXmlRoot();
			}
			@Override
			protected boolean dataUsesAttributes(){
				return false?false:super.dataUsesAttributes();
			}
			public XmlSpecifier[]fileSpecifiers(){
				return false?new XmlSpecifier[]{
					new XmlSpecifier("state.xml","Facets state files",this),
					super.fileSpecifiers()[0],
					super.fileSpecifiers()[1],
				}
				:super.fileSpecifiers();
			};
		};
	}
	/**
	Fulfils {@link FacetAppSurface#getInternalContentSource()} in the
	return of {@link #newApp(FacetFactory)}.
	@return by default {@link Nodes#newTestTree(String, int, double)} with <code>broad</code>=5 which
	may be changed by passing {@link #ARG_TREE_SIZE}=<code>broad</code> 
	 */
	protected Object getInternalContentSource(){
		return Nodes.newTestTree("Test",
				values.nature(false).getOrPutInt(ARG_TREE_SIZE,5),0.75);
	}
	/**
	Enables redefinition of {@link SView}s supplying viewer policy. 
	<p>Called from {@link TreeAppContenter#newContentViewers(ViewableFrame)}. 
	@param viewable has the content as its {@link SFrameTarget#framed}
	@return by default a single {@link TreeView}
	 */
	protected SView[]newContentViews(NodeViewable viewable){
		final String rootTitle=((TypedNode)viewable.framed).title();
		final boolean liveViews=canEditContent();
		SView view=new TreeView("View"){
			public boolean showRoot(){
				return !rootTitle.endsWith(DataConstants.TYPE_XML);
			}
			public boolean isLive(){
				return liveViews;
			}
		},
		basic=new TreeView("Basic"){
			@Override
			public String contentIconKey(Object content){
				return null;
			}
			public boolean isLive(){
				return liveViews;
			}
		};
		return false?new SView[]{basic,view}:new SView[]{view};
	}
	/**
	Enables definition of additional {@link STarget}s to be exposed in the
	{@link TreeTargets#TITLE_MENU} menu. 
	<p>Called from {@link TreeAppContenter#lazyContentRootElements(SAreaTarget)}. 
	@return targets whose {@link STargeter}s will be  
	passed to {@link #newTreeMenuItems(FacetFactory, STargeter[], STargeter[])};
	by default an empty {@link STarget}<code>[]</code>
	 */
	protected STarget[]newContentRootTargets(){
		return new STarget[]{};
	}
	/**
	Defines {@link SFacet} items for the {@link TreeTargets#TITLE_MENU} menu. 
	<p>Called from {@link TreeAppFacets#adjustMenuRoots(MenuFacets[])} 
	@param f the calling {@link TreeAppFacets} 
	@param treeLinks exposing {@link TreeTargets#appTargets()} 
	@param contentLinks targeting returns (if any) of {@link #newContentRootTargets()} 
	@return by default <code>treeAppFacets</code> indexed with {@link TreeTargets#TARGET_SEARCH},
		{@link TreeTargets#TARGET_ENCODE},{@link TreeTargets#TARGET_TYPE},
	 */
	protected SFacet[]newTreeMenuItems(FacetFactory f,STargeter[]treeLinks,
			STargeter[]contentLinks){
		String hint=FacetFactory.HINT_NONE;
		return new SFacet[]{
			f.triggerMenuItems(treeLinks[TARGET_SEARCH],hint),
			f.triggerMenuItems(treeLinks[TARGET_TYPE],hint),
			f.triggerMenu(treeLinks[TARGET_ENCODE],hint),
		};
	}
}/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
