package facets.facet.toolkit.swing;
import static java.lang.Math.round;
import facets.util.Debug;
import facets.util.Util;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.List;
import javax.swing.FocusManager;
import javax.swing.Icon;
import javax.swing.JComponent;
import javax.swing.JDesktopPane;
import javax.swing.JInternalFrame;
import javax.swing.JPanel;
import javax.swing.JInternalFrame.JDesktopIcon;
import javax.swing.JRootPane;
import javax.swing.SwingUtilities;
/**
Manages content windows within an internal desktop. 
<p>{@link AppDesktop} offers an alternative to {@link JDesktopPane} by encapsulating the 
essential contract of an internal desktop. 
<p>Content is added by means of {@link ContentPanel}s; 
the desktop base is returned by {@link #getPanel()}. 
<h4>Design</h4>
<ul>
<li>Content frames are maintained inside the {@link AppDesktop} independently of 
the {@link JDesktopPane} that displays them. 
<li> Content is adjusted by passing a new set of panels to {@link #setContents(Iterable)}. 
<li>Content frames are set in the the {@link JDesktopPane} by {@link #updateLayout(Sizes)} which then retiles all frames.  
<li>Frame layout is finally set by {@link #adjustToSizes(Sizes)}. 
</ul>
<h4>Functional test</h4>
<p>Cycle frame states after every step. 
<ol>
<li>Open app. 
<li>New window. 
<li>Minimise active,restore
<li>Close last window, open new one. 
</ol>
 */
final class AppDesktop extends JDesktopPane{
	/**
		Container for content to be added to a {@link AppDesktop}. 
		<p>{@link ContentPanel} extends its superclass by defining abstract methods which
		enable communication between the {@link AppDesktop} and its client application.
		<p>All methods are callbacks used by the {@link AppDesktop} to communicate with the client;
		client code should 
	<ul><li>use methods from {@link JPanel} to add content, which will be displayed by
	the {@link AppDesktop} in a window. 
	    <li>extend the contract as required to enable content already added to be identified	</ul>	 
		*/
	public static abstract class ContentPanel extends JPanel{
		/**
		Supply a title for the window frame. 
		 */
		protected abstract String windowTitle();
		/**
		Can the content window be safely closed? 
		<p><b>Note</b> Closing the window does not affect the references returned by
		{@link AppDesktop#getContents()}. 
		@return <code>false</code> to abort the close. 
		 */
		public abstract boolean windowCanClose();
		/**
		The content window has been maximised or restored. 
		<p>Most useful as a proxy for the state of the {@link AppDesktop}.  
		@param onOff the state of the content window 
		 */
		protected abstract void windowMaximumSet(boolean onOff);
		/**
		The content window has been activated. 
		 */
		protected abstract void windowActivated();
		protected abstract Icon windowIcon();
	}
	final static class Sizes extends Dimension{
		int iconGap;
		Sizes(Dimension size,int iconGap){
			super(size);
			this.iconGap=iconGap;
		}
		int heightLessGap(){
			return height-iconGap;
		}
		public String toString(){
			return "[" +height+","+iconGap+"]";
		}
	}
	boolean settingActive;
	private final JComponent panel;
	final List<AppDesktopFrame>frames=new ArrayList();
	private boolean scaleFrames;
	Sizes sizes=new Sizes(new Dimension(),0);
	boolean framesMaximum;
  private void trace(String msg,Object o){
		if(false)Util.println(getClass().getSimpleName()+msg,o);
	}
	/**
  Unique constructor. 
  @param background the background colour for the desktop
  @param defaultMaximum should the first window be opened maximised?
  @param scaleToDesktop equivalent to calling {@link #setScaleToDesktop(boolean)}   
   */
  AppDesktop(Color background,boolean defaultMaximum,boolean scaleToDesktop){
		setBackground(background);
		framesMaximum=defaultMaximum;
		scaleFrames=scaleToDesktop;
		panel=new JPanel(new BorderLayout());
		panel.add(this,BorderLayout.CENTER);
		if(false)addMouseMotionListener(new MouseAdapter(){
			public void mouseMoved(MouseEvent e){
				trace(": e=",e);
			}
		});
	}
  /**
  Sets panels to be displayed in desktop windows. 
  <p>Each call must be followed by a call to {@link #setActiveContent(ContentPanel)}; 
  underlying window sizes are set on each call as if by a call to {@link #tileWindows()};  
	<p>The method is also used to adjust content by passing an appropriately modified
	(even empty) <code>panels</code>
	  @param panels will be returned by {@link #getContents()}
   */
  public void setContents(Iterable<ContentPanel>panels){
  	if(panels==null)throw new IllegalArgumentException(
  			"Null panels in "+Debug.info(this));
		List<AppDesktopFrame>nowFrames=new ArrayList();
		for(ContentPanel panel:panels){
			AppDesktopFrame addFrame=null;
			for(AppDesktopFrame thenFrame:frames)
				if(thenFrame.content==panel)nowFrames.add(addFrame=thenFrame);
			if(addFrame==null)
				nowFrames.add(addFrame=new AppDesktopFrame(this,panel));
		}
		frames.clear();
		frames.addAll(nowFrames);
		if(frames.size()==0)sizes.iconGap=0;
		trace(".setContents: frames=",Util.arrayPrintString(frames.toArray()));
		removeAll();
		updateLayout(sizes);
	}
	/**
	Set the (possibly iconised) window to be displayed as active. 
	@param panel contained by the window to be activated
	 */
	public void setActiveContent(ContentPanel panel){
		if(panel==null)throw new IllegalStateException(
				"Null panel in "+Debug.info(this));
		trace(".setActiveContent: selected frame=",getSelectedFrame());
		settingActive=true;
		for(AppDesktopFrame frame:frames){
			frame.updateTitle();
			frame.setVisible(true);
			if(frame.framesContent(panel)){
				trace(".setActiveContent: setting frame=",frame);
				frame.trySetSelected(true);
			}
		}
		settingActive=false;
	}
	private AppDesktopFrame singleFrame(){
		return frames.size()==1?frames.get(0):null;
	}
	/**
	Re-implementation. 
	<p>Returns immediately if panel has zero height, 
	otherwise calls {@link #updateLayout(Sizes)} with current size. 
	 */
	public void doLayout(){
		Sizes nowSizes=new Sizes(getSize(),sizes.iconGap);
		if(false)trace(".doLayout: nowSizes=",nowSizes);
		if(nowSizes.height==0)return;
		updateLayout(nowSizes);
	}
	/**
	Main updating method. 
<ol>
	<li>Checks for single content 
	<li>Where content has changed, updates frames and tiles with {@link #tileToSizes(Sizes)} 
	<li>Activates and maybe maximises single (eg first) frame 
	<li>If size has changed, calls {@link #adjustToSizes(Sizes)} and records 
	current size. 
	</ol>
		<p>Called from 
<ul>
	<li>{@link #doLayout()} with non-zero current size
	<li>{@link #setContents(Iterable)} with last recorded size
	</ul>
	@param nowSizes current size
	 */
	private void updateLayout(Sizes nowSizes){
		if(false)trace(".updateLayout: nowSizes=",nowSizes);
		if(!nowSizes.equals(getSize()))throw new IllegalStateException(
				"Bad nowSizes="+nowSizes);
		AppDesktopFrame single=singleFrame();
		boolean settingContent=getComponentCount()==0;
		if(single!=null)nowSizes.iconGap=0;
		if(settingContent){
			for(AppDesktopFrame frame:frames)
				add(!frame.isIcon()?frame:frame.getDesktopIcon());
			tileToSizes(nowSizes);
		}
		if(single!=null){
			if(single.getSize().height<=0)tileToSizes(nowSizes);
			AppDesktopFrame frame=frames.get(0);
			frame.trySetSelected(true);
			frame.trySetMaximum(framesMaximum);
		}
		if(true)trace(".updateLayout: frames=",frames);
		repaint();
		if(frames.size()==0)SwingUtilities.invokeLater(new Runnable(){public void run(){
			JRootPane rootPane=getRootPane();
			if(rootPane!=null)rootPane.requestFocusInWindow();
		}});
		if(nowSizes.equals(sizes))return;
		adjustToSizes(nowSizes);
  	sizes=nowSizes;
	}
	/**
	Sets tiled restore sizes for all non-iconified frames.
<ol>
	<li>Calculates tiled sizes assuming no icon gap
	<li>Calls {@link #adjustToSizes(Sizes)} to adjust results for any gap 
	</ol>
		<p>Called from
<ul>
	<li>{@link #updateLayout(Sizes)} where content has changed
	<li>Client call of {@link #tileWindows()} 
	</ul>
		@param nowSizes current size
	 */
	private void tileToSizes(Sizes nowSizes){
		trace(".tileToSizes: nowSizes=",nowSizes);
		if(!nowSizes.equals(getSize()))throw new IllegalStateException(
				"Bad nowSizes="+nowSizes);
		List<AppDesktopFrame>tileFrames=new ArrayList();
		for(AppDesktopFrame frame:frames)
			if(!frame.isIcon())tileFrames.add(frame);
		int splitLong,splitShort,tileCount=tileFrames.size();
		switch(tileCount){
		case 0:return;
		case 1:splitLong=1;splitShort=1;break;
		case 2:splitLong=2;splitShort=1;break;
		case 3:case 4:splitLong=2;splitShort=2;break;
		case 5:case 6:splitLong=3;splitShort=2;break;
		case 7:case 8:case 9:splitLong=3;splitShort=3;break;
		case 10:case 11:case 12:splitLong=4;splitShort=3;break;
		case 13:case 14:case 15:case 16:splitLong=4;splitShort=4;break;
		default:
			throw new IllegalStateException("Bad 16+ frames in "+Debug.info(this));
		};
		boolean wide=nowSizes.width>nowSizes.height;
		double width=nowSizes.width/(wide?splitLong:splitShort),
			height=(nowSizes.height-nowSizes.iconGap)
				/(!wide?splitLong:splitShort);
		for(int at=0;at<tileCount;at++){
			int divisor=wide?splitLong:splitShort,
				row=at/divisor,col=at%divisor;
			double useWidth=width*(at<tileCount-1?1:splitLong*splitShort-at);
			AppDesktopFrame frameAt=tileFrames.get(at);
			frameAt.setRestoreBounds(new Rectangle(
					(int)(col*width),(int)(row*height),(int)useWidth,(int)height));
		}
		if(true||sizes.iconGap>0)adjustToSizes(nowSizes);
	}
	/**
	Adjusts desktop elements to current size and icon gap. 
	<p>For each frame,
<ul>
	<li>if iconified, scales the icon location to the current size
	<li>otherwise, if scaling is on or icon gap has changed, adjusts size for current gap 
	</ul>
		<p>Called by 
<ul>
	<li>{@link #tileToSizes(Sizes)} if there is an icon gap
	<li>{@link #updateLayout(Sizes)} if size has changed 
	<li>the first {@link AppDesktopFrame} to be iconified
	</ul>
	@param nowSizes current size
	 */
	void adjustToSizes(Sizes nowSizes){
		if(sizes.height==0)return;
		trace(".adjustToSizes: nowSizes=",nowSizes);
		if(!nowSizes.equals(getSize()))throw new IllegalStateException(
				"Bad nowSizes="+nowSizes);
		final double ratioW=(double)nowSizes.width/sizes.width,
			ratioH=(double)nowSizes.height/sizes.height,
			ratioHGap=(double)nowSizes.heightLessGap()/sizes.heightLessGap();
		for(AppDesktopFrame frame:frames){
			if(frame.isIcon()){
				JComponent icon=frame.getDesktopIcon();
				Rectangle bounds=icon.getBounds();
				int newX=(int)(bounds.x*ratioW),newY=(int)(bounds.y*ratioH);
				boolean atBottom=bounds.y+bounds.height>=sizes.height||
					newY+bounds.height>=nowSizes.height;
				icon.setLocation(newX,!atBottom?newY:nowSizes.height-bounds.height);
				trace(".adjustToSizes: ",icon);
			}
			if(scaleFrames||nowSizes.iconGap!=sizes.iconGap){
				boolean debug=true;
				if(debug)trace(".doLayout: scaling for sizes=", sizes+
						" nowSizes="+nowSizes);
				Rectangle bounds=frame.getBounds();
				frame.setRestoreBounds(new Rectangle(
						(int)round(bounds.x*ratioW),(int)round(bounds.y*ratioHGap),
						(int)round(bounds.width*ratioW),(int)round(bounds.height*ratioHGap)
					));
				if(debug)trace(".adjustToSizes: ",frame);
			}
		}
	}
	 /**
	Restores, resizes and arranges all non-iconified content windows. 
	<p>Once a window has been iconified, windows are tiled above
	a 'gap' to avoid obscuring icons; the gap is reset to zero if all windows are 
	closed.    
	 */
	public void tileWindows(){
		tileToSizes(sizes);
		((AppDesktopFrame)getSelectedFrame()).trySetMaximum(false);
	}
	/**
	Defines the behaviour of the {@link AppDesktop} on resize.
	<ul>
	<li><b>Off</b> - Window icon locations scaled to  desktop, 
	active windows unaffected.    
	<li><b>On</b> - Icons rearranged, active windows resized to the new desktop dimensions. 
	</ul> 
	@param onOff sets the behaviour
	 */
	public void setScaleToDesktop(boolean onOff){
		scaleFrames=onOff;
	}
	/**
	  Get contents currently displayed in desktop windows. 
	@return contents last passed to {@link #setContents(Iterable)} 
	 */
	public Iterable<ContentPanel>getContents(){
		List<ContentPanel>contents=new ArrayList();
		for(AppDesktopFrame frame:frames)contents.add(frame.content);
		return contents;
	}
	/**
	The desktop base. 
	 */
	public JComponent getPanel(){
		return panel;
	}
}/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
