package facets.util;
import static facets.util.Debug.*;
import facets.util.Util.SerializableBytes;
import java.io.Serializable;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;
/**
Converts between {@link Serializable}s and compressed bytes.
 */
final public class PackedSerialized extends Tracer implements SizeEstimable{
	private static final boolean packing=true,verbose=false;
	protected void traceOutput(String msg){
		msg=" "+(hashCode()&0xfff)+msg;
		if(false&&verbose)Times.printElapsed(Debug.info(this)+msg);
		else if(false)super.traceOutput(msg);
	}
	/**
	The result of passing a {@link Serializable} to the constructor. 
	 */
	public final byte[]packed;
	private final int unpackedLength;
	@Override
	public String toString(){
		return Debug.info(this)+" unpackedLength="+unpackedLength+" packed="+packed.length;
	}
	/**
	Unique constructor. 
	@param serializable will be converted to a compressed <code>byte[]</code> 
	stored as {@link #packed}
	 */
	public PackedSerialized(Serializable serializable){
		this(serializable,packing);
	}
	private PackedSerialized(Serializable serializable,boolean doPack){
		Debug.mbAsKb=true;
		if(verbose)traceDebug(": serializable=",serializable);
		byte[]bytes=SerializableBytes.serialize(serializable);
		unpackedLength=bytes.length;
		packed=doPack?packBytes(bytes):bytes;
		trace(": unpacked="+mb(unpackedLength)+" packed="+mb(packed.length)+" %="+packPct());
		Debug.mbAsKb=false;
	}
	private PackedSerialized(byte[]packed){
		this.packed=packed;
		unpackedLength=0;
	}
	/**
	Attempts to recreate the {@link Serializable} passed to the constructor. 
	<p>Throws a {@link RuntimeException} if recreation fails. 
	 */
	public Serializable unpack(){
		Debug.mbAsKb=true;
		Serializable deserialized=null;
		byte[]unpacked=null;
		if(packing)try{
			unpacked=unpackBytes(packed,unpackedLength);
		}catch(DataFormatException e){
			if(verbose)trace(".unpack: e=",e);
		}
		if(unpacked==null)unpacked=packed;
		deserialized=SerializableBytes.deserialize(unpacked);
		trace(".unpack: packed="+mb(packed.length)+
				" unpacked="+mb(unpacked.length)+" %="+packPct());
		if(verbose)traceDebug(".unpack: deserialized=",deserialized);
		Debug.mbAsKb=false;
		return deserialized;
	}
	/**
	Creates a {@link PackedSerialized} that is guaranteed to unpack. 
	@param s passed to constructor
	@return a {@link PackedSerialized} guaranteed to unpack but possibly not
	compressed  
	 */
	public static PackedSerialized newUnpackChecked(Serializable s){
		PackedSerialized ps;
		try{
			ps=new PackedSerialized(s,packing);
			ps.unpack();
			return ps;
		}catch(Exception e){
			if(verbose)Util.println("PackedSerialized.newUnpackChecked e=",e);
			ps=new PackedSerialized(s,false);
			ps.unpack();
			return ps;
		}
	}
	/**
	Attempts to recreate a {@link Serializable} from bytes. 
	@param packed is treated as a valid {@link #packed}
	@return a {@link Serializable} or <code>null</code> if recreation fails 
	 */
	public static Serializable newUnpacked(byte[]packed){
		try{
			return new PackedSerialized(packed).unpack();
		}catch(Exception e){
			if(verbose)Util.println("PackedSerialized.newUnpacked e=",e);
			return null;
		}
	}
	@Override
	public long estimateSize(){
		if(packed==null)throw new IllegalStateException("Null packed in "+Debug.info(this));
		return packed.length;
	}
	private int packPct(){
		return unpackedLength<=0?-1:100-Util.intPct(packed.length,unpackedLength);
	}
	private static byte[]packBytes(byte[]in){
		byte[]out=new byte[in.length];
		Deflater zip=new Deflater();
		zip.setInput(in);
		zip.finish();
		int count=zip.deflate(out);
		return truncatedBytes(out,count);
	}
	private static byte[]unpackBytes(byte[]in,int lengthHint)
			throws DataFormatException{
		byte[]out=new byte[lengthHint<=0?in.length*100:lengthHint*11/10];
		Inflater zip=new Inflater();
		zip.setInput(in);
		int count=zip.inflate(out);
		return truncatedBytes(out,count);
	}
	private static byte[]truncatedBytes(byte[]in,int count){
		byte[]out=new byte[count];
		System.arraycopy(in,0,out,0,count);
		return out;
	}
}/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
