package facets.util.app;
import static facets.core.app.AppConstants.*;
import static facets.util.app.AppValues.*;
import static facets.util.datatree.Nodes.*;
import facets.util.Debug;
import facets.util.FileSpecifier;
import facets.util.TracerInput;
import facets.util.datatree.DataNode;
import facets.util.datatree.TypedNode;
import facets.util.datatree.ValueNode;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
/**
Manages file history and other file values. 
<p>{@link AppFileValues} encapsulates file-related operations that read and write 
values stored in an {@link AppValues}.  
 */
public abstract class AppFileValues{
	/**
	Key for flag specifying file behaviour. 
	 */
	public static final String 
		NATURE_FILES="Files",
		NATURE_FILES_RECENT_MAX="maximumRecent",
		NATURE_FILES_OPENING="openingFile",
		NATURE_FILES_OPENING_PREVIOUS="previous";
	private final AppValues appValues;
	private final ValueNode natureFiles;
	private final int maxRecent;
	private File[]recentFiles;
	/**
	Unique constructor. 
	@param appValues stores values. 
	 */
	public AppFileValues(AppValues appValues){
		if(!appValues.hasSystemAccess())throw new IllegalStateException(
				"No system access in in "+Debug.info(this));
		else this.appValues=appValues;
		natureFiles=guaranteedValueChild(appValues.nature(true),NATURE_FILES);
		maxRecent=natureFiles.getOrPutInt(NATURE_FILES_RECENT_MAX,9);
	}
	/**
	Returns a validated array of recently opened files.
	@return files updated by {@link #updateRecentFiles(File)}
	 */
	final public File[]recentFiles(){
		return recentFiles;
	}
	/**
	Maintains the list returned by {@link #recentFiles()}. 
	<p>Files are guaranteed to exist, and are returned in reverse order of opening.  
	@param add a file to add, or <code>null</code> to test the values stored
	@return the updated {@link ValueNode}
	 */
	final public ValueNode updateRecentFiles(File add){
		ValueNode recent=appValues.state(PATH_CORE,STATE_NODE_FILES,
				STATE_NODE_RECENT);
		String[]existing=recent.values();
		List<File>tested=new ArrayList();
		if(add!=null&&add.isFile()){
			tested.add(add);
			statePutPath(add.getParentFile());
		}
		for(int i=0,count=tested.size();i<existing.length;i++){
			File test=new File(existing[i]);
			if(!test.isFile()||test.equals(add)||count++==maxRecent)continue;
			else tested.add(test);
		}
		recent.setValues(recentFiles=tested.toArray(new File[]{}));
		return recent;
	}
	/**
	Tries to get a suitable opening content file. 
	@param mustGet <code>null</code> return not allowed
	@return the file or <code>null</code> unless <code>mustGet</code> is <code>true</code>,
	in which case calls {@link #gotNoOpening()}
	 */
	final public File getOpeningFile(boolean mustGet){
		DataNode recent=updateRecentFiles(null);
		String[]recentPaths=recent.values();
		FileSpecifier[]specifiers=getOpenSpecifiers();
		File got=recentPaths.length>0?getOpeningRecent(recentPaths)
				:getOpeningOther(specifiers);
		if(got==null||!got.isFile()){
			if(!mustGet)return got;
			else gotNoOpening();
		}
		else updateRecentFiles(got);
		return got;
	}
	/**
	Get {@link FileSpecifier}s specifying application content types. 
	@return by default {@link FileSpecifier#ALL} 
	 */
	public FileSpecifier[]getOpenSpecifiers(){
		return new FileSpecifier[]{FileSpecifier.ALL};
	}
	/**
	Called from {@link #getOpeningFile(boolean)}. 
	 */
	protected File getOpeningRecent(String[]recentPaths){
		throw new RuntimeException("Not implemented in "+Debug.info(this));
	}
	/**
	Called from {@link #getOpeningFile(boolean)}. 
	 */
	protected File getOpeningOther(FileSpecifier[]specifiers){
		throw new RuntimeException("Not implemented in "+Debug.info(this));
	}
	/**
	Called from {@link #getOpeningFile(boolean)}. 
	 */
	protected void gotNoOpening(){
		throw new RuntimeException("Not implemented in "+Debug.info(this));
	}
	/**
	Gets the current file search path. 
	@return the last valid file passed to {@link #statePutPath(File)}, or else 
	(for Windows) the user desktop. 
	 */
	final public File stateGetPath(){
		String path=stateFiles().getString(STATE_FILES_PATH);
		File file=new File(path).getAbsoluteFile();
		if(path.equals("")||!file.isDirectory())
			statePutPath(file=new File(System.getProperty("user.home"),"Desktop"
			).getAbsoluteFile());
		return file;
	}
	/**
	Stores the path to be returned by {@link #stateGetPath()}. 
	@param file must be non-<code>null</code> and a directory
	 */
	final public void statePutPath(File file){
		if(file==null||!file.isDirectory())throw new IllegalArgumentException(
				"Null or non-directory path in "+Debug.info(this));
		else stateFiles().put(STATE_FILES_PATH,file);
	}
	private ValueNode stateFiles(){
		return appValues.state(PATH_CORE,STATE_NODE_FILES);
	}
	/**
	Creates an {@link AppFileValues} to read input from the console. 
	@param values for reading and writing
	@param s filter for file types to input
	 */
	public static AppFileValues newConsole(final AppValues values,final FileSpecifier s){
		return new AppFileValues(values){
			protected File getOpeningRecent(String[]recentPaths){
				String rubric="Recent files";
				Object got=TracerInput.getItemChoice(values,rubric,recentPaths);
				return got==null?null:new File((String)got).getAbsoluteFile();
			}
			protected File getOpeningOther(FileSpecifier[]specifiers){
				return TracerInput.getFile(values,stateGetPath(),specifiers[0]);
			}
			protected void gotNoOpening(){
				values.tryWrite("Exiting without file: ");
				System.exit(-1);
			}
			public FileSpecifier[]getOpenSpecifiers(){
				return new FileSpecifier[]{s};
			}
		};
	}
	static void main(String[]args)throws Exception{
		AppValues values=new AppValues(AppFileValues.class){
			protected void traceOutput(String msg){
				if(false)traceOutputWithClass(msg);
			}
			protected void addNatureDefaults(ValueNode root){
				root.setContents(new ValueNode(PATH_CORE,TypedNode.UNTITLED,
					new Object[]{
					NATURE_WRITABLE+"="+false,
					NATURE_RECORD_RUNS+"="+true,
				}));
			}
		};
		values.readValues(args);
		values.trace("Got file ",newConsole(values,FileSpecifier.ALL
			).getOpeningFile(true));
		values.tryWrite("");
	}
}/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
