package facets.util.app;
import static facets.util.app.AppValues.*;
import static java.lang.System.*;
import static java.lang.Thread.*;
import facets.util.Debug;
import facets.util.Objects;
import facets.util.Times;
import facets.util.Tracer;
import facets.util.Util;
import facets.util.app.WatchableOperation.CancelStyle;
import facets.util.app.AppWatcher.WatcherCoupler;
import java.awt.EventQueue;
import java.util.ArrayList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
/** High-level monitor of application program operation. 
<p>{@link AppWatcher} has three closely-related functions. 
<ul>
	<li><b>Busy checking</b>: Detecting blocking of the event loop. 
	<li><b>Watching</b>: Handling of exceptions thrown by {@link WatchableOperation}s. 
	<li><b>Cancellation</b>: Honouring {@link WatchableOperation#cancelStyle()}s. 
</ul>
<h3>Busy checking</h3> 
<p>By frequent calls to its (GUI toolkit dependent) implementation of {@link 
	#checkBusy()}, {@link AppWatcher} detects blocks in the event loop. This 
	enables it to 
<ul>
	<li>trigger changes in the current {@link BusyCursor} 
	<li>alert {@link WatcherCoupler} when an {@link WatchableOperation} requesting 
	{@link CancelStyle#Timeout} fails to complete within {@link #timeoutBlock} ms. 
	<li>trigger a system timeout if the event loop blocks for {@link #systemTimeout} 
		ms 
</ul>
<h3>Watching</h3>
<p>Exceptions thrown by {@link WatchableOperation}s are passed to {@link WatcherCoupler}. 
<h3>Cancellation</h3>
<p>While the {@link WatchableOperation} completes in a (high-priority) background 
	thread, client code can request cancellation as shown below. Cancellation 
	requests are passed to {@link WatcherCoupler} for confirmation. 
<table border="0" cellpadding="5">
	<tr> 
		<td width=30%><i>Style</i></td>
		<td><i>Enabled by</i></td>
		<td width=30%><i>Signalled by</i></td>
	</tr>
	<tr valign="top"> 
		<td>{@link CancelStyle#Dialog}</td>
		<td>Blocking the event loop by presenting a modal dialog</td>
		<td>Dialog cancelled</td>
	</tr>
	<tr valign="top"> 
		<td>{@link CancelStyle#Timeout} </td>
		<td>Blocking the event loop internally while checking for timeouts</td>
		<td>Timeout detected during busy checking</td>
	</tr>
	<tr valign="top"> 
		<td> {@link CancelStyle#Future}</td>
		<td>Restricting input, setting busy cursor and offering status message 
			via {@link WatcherCoupler}</td>
		<td>Input unrestricted</td>
	</tr>
</table>
*/
public abstract class AppWatcher extends Tracer{
	public static final String traceName=true?"AW":AppWatcher.class.getSimpleName();
	private final static int traceWait=500,busyWait=false?traceWait:50,
		debugTimeout=3,releaseTimeout=15;
	public static int timeoutBlock=debugTimeout*1000,systemTimeout=3*timeoutBlock;
	/**
		Enables client code to communicate with {@link AppWatcher}. 
		<p>The methods of {@link WatcherCoupler} enable client code to 
	<ul>
		<li>respond appropriately when a {@link WatchableOperation}  
		throws an exception 
		<li>tell {@link AppWatcher} whether or not actually to time operations
		that request this 
		<li>respond appropriately when either a {@link WatchableOperation} or the 
		application as a whole is timed out
		</ul>	 */
		public static class WatcherCoupler extends facets.util.Tracer{
			protected boolean retryCancel(List<WatchableOperation>ops){
				return false;
			}
			protected boolean checkTimeouts(){
				return true;
			}
			protected void handleException(List<WatchableOperation>ops,Exception e){
				throw new RuntimeException(ops.toString(),e);
			}
			protected void handleSystemTimeout(WatchableOperation op,Thread timed,
					boolean timing){
				System.out.println("System has been timed out in "+traceName+
					" after "+systemTimeout/1000+" sec:" +
					"\nwatched="+op+
					"\ntimed="+timed+
					"\nwaiting="+timing+
					"\nthreads="+Objects.toStringWithHeader(Debug.getSortedThreads()));
				exit(-1);
			}
			@Override
			protected void traceOutput(String msg){
				if(false)super.traceOutput(msg);
			}
			protected void operationStarting(List<WatchableOperation>ops){
				trace(".operationsStarting: ops=",ops);
			}
			protected void operationsCompleted(List<WatchableOperation>ops){
				trace(".operationsCompleted: ops=",ops);
			}
		}
	final protected void traceOutput(String msg){
		if(busyWait!=traceWait)return;
		msg=traceName+msg;
		if(false)System.out.println(msg);
		else Times.printElapsed(msg);
	}
	private static Object single;
	private static int calls=0;
	private final List<BusyCursor>cursors=new ArrayList();
	private WatcherCoupler coupler=new WatcherCoupler();
	private List<WatchableOperation>ops=new ArrayList();
	private long notBusyAt=currentTimeMillis();
	protected Thread blocking;
	private boolean passingException,timing;
	private Exception exception;
	private Object returnable;
	public AppWatcher(boolean inApplet){
		if(single!=null)throw new IllegalStateException(
				"Only one instance allowed in "+Debug.info(this));
		else single=this;
		Thread[]threads=Debug.getSortedThreads();
		if(threads.length>(inApplet?12:8))throw new IllegalStateException(
				threads.length+" starting threads in "+Debug.info(this));
		else if(false)Util.println(".AppWatcher: threads=",threads);
		timeoutBlock=(Debug.natureDebug?debugTimeout:releaseTimeout)*1000;
		systemTimeout=3*timeoutBlock;
		if(!Debug.natureDebug&&busyWait==traceWait)throw new IllegalStateException(
				"busyWait==traceWait in "+Debug.info(this));
		else if(false)Util.println("AppWatcher timeout=",timeoutBlock);
		final String taskTitle=traceName+".checks";
		new Timer(taskTitle).schedule(new TimerTask(){
			public String toString(){
				return taskTitle;
			}
			public void run(){
				boolean busy=checkBusy();
				if(cursors.size()>0)cursors.get(0).checkBusyState(busy);
				long nowAt=currentTimeMillis();
				if(!busy){
					notBusyAt=nowAt;
					return;
				}
				long busyAt=nowAt-notBusyAt;
				if(timing)trace(".checks: busyAt="+busyAt/1000+" "+ops.get(0));
				if(systemTimeout>0&&busyAt>systemTimeout&&ops.size()>0)
					coupler.handleSystemTimeout(ops.get(0),blocking,timing);
				else if(busyAt<timeoutBlock||!timing||exception!=null)return;
				trace(".checks: busyAt="+busyAt/1000+" "+blocking);
				timing=false;
			}
		},0,busyWait);
	}
	/**
	Runs (and possibly allows cancellation of) an {@link WatchableOperation}. 
	<p>Exceptions thrown are passed to the {@link WatcherCoupler} returned by {@link #coupler}.  
	@param op may be created within another {@link WatchableOperation} providing
	this specifies {@link CancelStyle#None}. 
	@return <code>null</code> or the value returned by {@link WatchableOperation#doOperations(int)}
	 */
	public final Object runWatched(final WatchableOperation op){
		returnable=null;
		final Thread callThread=currentThread();
		int callPriority=callThread.getPriority();
		final int call=++calls%100;
		CancelStyle cancelStyle=op.cancelStyle();
		trace(".runWatched: @" +call+" "+op+" cancelStyle="+cancelStyle);
		try{
			if(!EventQueue.isDispatchThread())throw new IllegalStateException(
					"Not event despatch thread "+callThread);
			else if(cancelStyle==CancelStyle.Dialog||
					(cancelStyle==CancelStyle.Timeout&&(false||coupler.checkTimeouts()))){
				if(blocking!=null)throw new IllegalStateException(
						op+" cannot be cancelled: active blocking="+blocking);
				else if(false&&busyWait==traceWait)Runtime.getRuntime().gc();
				blocking=new Thread(traceName+".blocking: "+op){
					public void interrupt(){
						throw new RuntimeException(this.toString());
					};
					public void run(){		
						try{
							doOperations(op,call);
						}catch(Exception e){
							exception=e;
						}
						trace(".releaseBlock: caller=",this);
						blocking=null;
						timing=false;
						prioritiseThreads(callThread,blocking);
						cancelBlockingDialog();
					}
				};
				blocking.start();
				while(true){
					trace(".doBlock @"+call +" cancelStyle="+cancelStyle);
					if(cancelStyle==CancelStyle.Dialog){
						if(false&&cursors.size()>0)cursors.get(0).forceSwitch(true);
						openBlockingDialog(op);
					}
					else{
						notBusyAt=currentTimeMillis();
						timing=true;
					}
					prioritiseThreads(blocking,callThread);
					while(timing);
					prioritiseThreads(callThread,blocking);
					if(exception!=null||blocking==null||!blocking.isAlive()
						||!coupler.retryCancel(ops))break;
				}
				if(blocking!=null&&blocking.isAlive())blocking.setPriority(MIN_PRIORITY);
				blocking=null;
			}
			else if(cancelStyle==CancelStyle.Future){
				scheduleFuture(op);
				return null;
			}
			else doOperations(op,call);
		}catch(Exception e){
			exception=e;
		}
		if(exception!=null){
			if(ops.size()==1||blocking!=null||passingException){
				Throwable t=passingException&&exception.getCause()!=null?exception.getCause()
						:exception;
				coupler.handleException(ops,(Exception)t);
				passingException=false;
				exception=null;
			}else{
				passingException=true;
				throw false?new RuntimeException(
						"Ops="+Objects.toStringWithHeader(ops.toArray()),exception)
					:new RuntimeException(exception);
			}
		}
		else{
			if(op!=ops.get(0))throw new IllegalStateException(
					"Invalid ops "+ops.size());
			coupler.operationsCompleted(ops);
			ops.remove(0);
		}
		trace(".~runWatched @"+call+" returnable="+Debug.info(returnable)
				+" exception="+exception);
		callThread.setPriority(callPriority);
		notBusyAt=currentTimeMillis();
		return returnable;
	}
	private void doOperations(WatchableOperation op,int call){
		ops.add(0,op);
		if(false)trace(".doOperations: @"+call+" ops="+ops.size()+" blocking="+blocking);
		coupler.operationStarting(ops);
		returnable=op.doOperations(call);
		if(false)traceDebug(".~doOperations: @"+call+" returnable=",returnable);
	}
	private void prioritiseThreads(Thread up,Thread down){
		boolean complain=false;
		if(complain&&up==null)throw new IllegalArgumentException("Null up for down="+down);
		else if(complain&&down==null)throw new IllegalArgumentException("Null down for up="+up);
		if(up!=null)up.setPriority(MAX_PRIORITY);
		if(down!=null)down.setPriority(MIN_PRIORITY);
		trace(".prioritiseThreads: up="+up+" down="+down);
	}
	protected void openBlockingDialog(WatchableOperation op){}
	protected void cancelBlockingDialog(){}
	protected boolean checkBusy(){
		return false;
	}
	protected void scheduleFuture(WatchableOperation op){
		throw new RuntimeException("Not implemented in "+Debug.info(this));
	}
	final protected void runFuture(WatchableOperation op){
		futureRunning(true);
		futureRunning(false);
	}	
	protected void futureRunning(boolean on){
		throw new RuntimeException("Not implemented in "+Debug.info(this));
	}
	final public void setCoupler(WatcherCoupler coupler){
		if(coupler==null)throw new IllegalArgumentException("Null coupler in "+Debug.info(this));
		else this.coupler=coupler;
	}
	final public void pushCursor(BusyCursor cursor){
		cursors.add(0,cursor);
	}
	final public void popCursor(){
		cursors.remove(0);
		if(cursors.size()>1)throw new IllegalStateException(
				"Empty cursors in "+Debug.info(this));
	}
}/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
