package facets.util.datatree;
import static facets.util.datatree.Nodes.*;
import facets.util.Debug;
import facets.util.Objects;
import facets.util.Util;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
/**
{@link List} wrapping a {@link DataNode} and its {@link TypedNode#children()}. 
 */
public final class NodeList extends ArrayList<TypedNode>{
	/** The node whose children the {@link NodeList} is manipulating. */
	public final DataNode parent;
	private final boolean autoUpdate;
	/**
	Unique constructor. 
	@param parent must be a {@link DataNode} whose children the {@link NodeList} will 
	manipulate
	@param autoUpdate should each operation of the {@link NodeList} 
	be reflected immediately in  {@link #parent}? 
	 */
	public NodeList(TypedNode parent,boolean autoUpdate){
		this.parent=(DataNode)parent;
		this.autoUpdate=autoUpdate;
		super.addAll(Arrays.asList(parent.children()));
	}
	/**
	Update {@link #parent} with the contents of the {@link NodeList}. 
	<p>Called internally on each operation unless <code>false</code> was passed
	to {@link NodeList#NodeList(TypedNode, boolean)}, in which case client code
	must call as appropriate. 
	 */
	public void updateParent(){
		if(hasMixedContents(parent))throw new IllegalArgumentException(
				"Mixed contents in parent="+parent);
		parent.setChildren(toArray(new TypedNode[]{}));
	}
	public boolean add(TypedNode n){
		boolean added=super.add(n);
		if(autoUpdate)updateParent();
		return added;
	}
	public TypedNode set(int at,TypedNode n){
		TypedNode set=super.set(at,n);
		if(autoUpdate)updateParent();
		return set;
	}
	public void add(int at,TypedNode n){
		super.add(at,n);
		if(autoUpdate)updateParent();
	}
	public boolean addAll(Collection<?extends TypedNode>c){
		boolean all=super.addAll(c);
		if(autoUpdate)updateParent();
		return all;
	}
	public boolean remove(Object o){
		boolean removed=super.remove(o);
		((TypedNode)o).setParent(null);
		if(autoUpdate)updateParent();
		return removed;
	}
	public TypedNode remove(int at){
		TypedNode remove=super.remove(at);
		if(autoUpdate)updateParent();
		return remove;
	}
	public boolean removeAll(Collection<?>c){
		boolean all=super.removeAll(c);
		if(autoUpdate)updateParent();
		return all;
	}
	public void clear(){
		super.clear();
		if(autoUpdate)updateParent();
	}
	public void addAll(TypedNode...nodes){
		addAll(Arrays.asList(nodes));
		if(autoUpdate)updateParent();
	}
	public void addAll(int at,TypedNode...nodes){
		addAll(at,Arrays.asList(nodes));
		if(autoUpdate)updateParent();
	}
	public void removeAll(TypedNode[]nodes){
		removeAll(Arrays.asList(nodes));
	}
	public TypedNode[]copyChildren(){
		updateParent();
		return((TypedNode)parent.copyState()).children();
	}
	public void updateMixedParent(Object insertMark){
		ArrayList update=new ArrayList();
		List contents=Arrays.asList(parent.contents());
		for(Object c:contents)
			if(c instanceof TypedNode){
				if(contains(c)){
					update.add(c);
					remove(c);
				}
			}
			else update.add(c);
		if(insertMark==null)update.addAll(this);
		else update.addAll(contents.indexOf(insertMark),this);
		if(false){
			Util.println("NodeList.updateMixedParent: parent=",parent.contents());
			Util.println("this=",toArray());
			Util.println("update=",update);
		}
		else parent.setContents(update.toArray());
	}
}/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
