package facets.util.datatree;
import static facets.util.ByteStrings.*;
import static facets.util.Objects.*;
import static facets.util.PackedSerialized.*;
import static facets.util.Regex.*;
import static facets.util.datatree.TypedNode.*;
import facets.util.Debug;
import facets.util.ItemList;
import facets.util.Objects;
import facets.util.Strings;
import facets.util.Tracer;
import facets.util.Util;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
/**
Utilities for manipulating and querying trees of {@link TypedNode}s and its subclasses. 
 */
public final class Nodes{
	public static final String NO_NULL="*";
	public static String encode(DataNode node,int breakAt){
		Object[]contents=node.contents();
		if(contents.length==0)return null;
		for(Object item:contents)if(item instanceof TypedNode)((TypedNode)item).setParent(null);
		String encoded=encodeBytes(newUnpackChecked(newTyped(Serializable.class,contents)).packed);
		if(encoded==null)return null;
		if(true)Util.println("Nodes.encode: "+encoded.length()+"\n"+encoded);
		node.setContents((breakAt<=0?encoded:encoded.replaceAll("(\\w{"+breakAt+"})","$1\n")
			).split("\n"));
		return encoded;
	}
	public static void decode(DataNode node){
		Object[]contents=node.contents();
		if(contents.length==0)return;
		for(Object object:contents)if(!(object instanceof String))return;
		byte[]packed=decodeString(Strings.linesString(newTyped(String.class,contents)));
		if(packed==null)return;
		Object[]unpacked=(Object[])newUnpacked(packed);
		if(unpacked!=null)node.setContents(unpacked);
	}
	public static String keySafe(String text){
		return text.trim().replaceAll("\\W","_");
	}
	public static boolean isKeyPair(String text){
		return contains(text,"^[\\w:]+=");
	}
	public static String[]splitPair(String text){
		return text.split("=");
	}
	/**
	Provides for (de)persisting the contents of a {@link DataNode}.   
		 */
	abstract static public class TreeRoot extends Tracer{
		/**
		Root node of the tree to read or write. 
		 */
		final public DataNode tree;
		/**
		Unique constructor. 
		@param node to (de)persist
		 */
		public TreeRoot(DataNode node){
			this.tree=node;
		}
		/**
		Read content into {@link #tree}. 
		*/
		public abstract void readFromSource(Object src)throws IOException;
		/**
		Write out content of {@link #tree}. 
		 */
		public abstract void writeToSink(Object sink)throws IOException;
	}
	public static String getSingleString(DataNode node){
		String[]strings=node.values();
		return strings.length==0?"":strings[0];
	}
	public static void setSingleString(DataNode node,String str){
		node.setContents(str);
	}
	/**
	Returns stringifications of the tree root and its descendants' 
	<code>type</code>, <code>title</code> and <code>values</code> 
	in contents order.
		 */
	public static String treeString(DataNode root){
		StringBuilder lines=new StringBuilder();
		for(TypedNode node:descendants(root)){
			lines.append("type="+node.type());
			lines.append("|title="+node.title());
			for(String each:((DataNode)node).values())lines.append("|"+each);
			lines.append("\n");
		}
		return lines.toString();
	}
	public static String childTypeText(TypedNode node,String type){
		TypedNode child=child(node,type);
		if(child==null)return "";
		Object[]values=child.values();
		return values.length==0?"":(String)values[0];
	}
	public static TypedNode child(TypedNode node,String type){
		TypedNode child=child(node,new String[]{type.replace(NO_NULL,"")},UNTITLED);
		if(type.endsWith(NO_NULL)&&child==null)throw new IllegalStateException(
				"No child of type '"+type+"' in "+node);
		return child;
	}
	public static TypedNode child(TypedNode node,String type,String title){
		return child(node,new String[]{type},title);
	}
	public static TypedNode child(TypedNode node,String[]types){
		return child(node,types,UNTITLED);
	}
	public static TypedNode child(TypedNode node,String[]types,String title){
	  TypedNode[]children=children(node,types);
	  for(int c=0;c<children.length;c++)
	  		if(title.equals(UNTITLED)||children[c].title().equals(title))
	  			return children[c];   	
	  return null;
	}
	public static TypedNode[]children(TypedNode node,String...types){
		if(node==null)throw new IllegalArgumentException("Null node");
	  TypedNode[]children=node.children();
		ItemList<TypedNode>matches=new ItemList(TypedNode.class);
	  for(int c=0;c<children.length;c++)for(int i=0;i<types.length;i++)
	  		if(children[c].type().equalsIgnoreCase(types[i]))
					matches.addItem(children[c]);
	  return matches.items();
	}
	public static TypedNode descendantTyped(TypedNode node,String type){
		for(TypedNode d:descendants(node))if(d.type().equals(type))return d;
		return null;
	}
	public static TypedNode descendantTitled(TypedNode node,String title){
		for(TypedNode d:descendants(node))if(d.title().equals(title))return d;
		return null;
	}
	public static TypedNode[]descendantsTyped(TypedNode node,String...types){
		List<TypedNode>list=new ArrayList();
		for(TypedNode d:descendants(node))for(String type:types)if(d.type().equals(type))list.add(d);
		return list.toArray(new TypedNode[]{});
	}
	public static TypedNode[]descendants(TypedNode node){
		if(node==null)throw new IllegalStateException("Null node");
		List<TypedNode>list=new ArrayList();
		list.add(node);
		for(TypedNode child:node.children())
		  list.addAll(Arrays.asList(descendants(child)));
		return list.toArray(new TypedNode[]{});
	}
	public static ValueNode guaranteedValueChild(DataNode parent,String type){
		return guaranteedChild(parent,type,UNTITLED);
	}
	public static ValueNode guaranteedChild(DataNode parent,
			String type,String title){
		ValueNode guaranteed=(ValueNode)child(parent,type,title);
		if(guaranteed!=null)return guaranteed;
		guaranteed=new ValueNode(type,title);
		appendChild(parent,guaranteed);
		return guaranteed;
	}
	public static ValueNode guaranteedDescendant(DataNode node,String...typePath){
		for(String type:typePath)node=guaranteedValueChild(node,type);
		return(ValueNode)node;
	}
	public static void setChild(DataNode parent,TypedNode child){
		List<TypedNode>children=new NodeList(parent,true);
		if(!children.contains(child))children.add(child);
		else for(TypedNode check:children)
			if(check.type().equals(child.type())&&check.title().equals(child.title()))
				children.set(children.indexOf(check),child);
	}
	public static void appendChild(DataNode node,TypedNode child){
		TypedNode[]children=node.children(),appended=new TypedNode[children.length+1];
		System.arraycopy(children,0,appended,0,children.length);
		appended[children.length]=child;
		node.setChildren(appended);
	}
	public static void appendValue(DataNode node,Object value){
		Object[]values=node.values(),appended=new Object[values.length+1];
		System.arraycopy(values,0,appended,0,values.length);
		appended[values.length]=value;
		node.setValues(appended);
	}
	public static TypedNode[]ancestry(TypedNode node){
		if(node.parent()==null)return new TypedNode[]{node};
		ItemList<TypedNode>ascent=new ItemList(TypedNode.class);
		TypedNode maybeTop=node.parent();
		ascent.addItems(new TypedNode[]{node,maybeTop});
		while(maybeTop.parent()!=null)
			ascent.addItem(maybeTop=maybeTop.parent());
		return(TypedNode[])Objects.reverse(ascent.items());
	}
	/**
	Merges new with existing node contents. 
	<p><code>mergeContents</code> will overwrite any values with the same keys
	@param node contains any existing contents
	@param mergeContents may include further {@link ValueNode}s which will be 
	merged with any of the same type, name and path
	@return <code>node</code> passed with merged contents
	 */
	public static ValueNode mergeContents(ValueNode node,Object...mergeContents){
		if(node==null)throw new IllegalArgumentException(
				"Null node for merge of "+Debug.info(mergeContents));
		if(mergeContents==null)throw new IllegalArgumentException(
				"Null mergeContents for "+Debug.info(node));
		for(int i=0;i<mergeContents.length;i++)
			if(mergeContents[i]instanceof String){
				String[]splits=((String)mergeContents[i]).split("=");
				if(splits.length==2)node.put(splits[0],splits[1]);
			}
		ItemList<TypedNode>mergedChildren=new ItemList(TypedNode.class);
		mergedChildren.addItems(node.children());
		TypedNode[]mergeNodes=new DataNode(
				DataConstants.TYPE_DATA,"Merge",mergeContents).children();
		for(int i=0;i<mergeNodes.length;i++){
			TypedNode nodeChild=child(node,
					mergeNodes[i].type(),mergeNodes[i].title());
			if(nodeChild!=null)mergeContents(
					(ValueNode)nodeChild,mergeNodes[i].contents());
			else mergedChildren.addItem(mergeNodes[i]);
		}
		node.setContents(node.values());
		node.setChildren(mergedChildren.items());
		return node;
	}
	public static String valuesFirstLine(TypedNode node){
		String[]lines=((DataNode)node).values();
		return lines[0]+(lines.length==1?"":"...");
	}
	public static String valuesAsLine(TypedNode node,int limit){
		String[]strings=((DataNode)node).values();
		StringBuffer b=new StringBuffer();
		for(int i=0;i<strings.length;i++){
			String s=strings[i];
			b.append(s+" ");			
		}
		String line=b.toString().replaceAll("\\s+"," ");
		return line.length()<limit?line:line.substring(0,limit)+"...";
	}
	public static String newValuesTitle(DataNode node){
		String[]values=node.values();
		if(values.length==0)return UNTITLED;
		String text=values[0];
		final int stop=20;
		return " ["+(text.length()>stop?text.substring(0,stop)+"...":text)+"]";
	}
	public static void pack(DataNode node){
		throw new RuntimeException("Not implemented for "+Debug.info(node));
	}
	public static void unpack(DataNode node){
		throw new RuntimeException("Not implemented for "+Debug.info(node));
	}
	/**
	@return a tree sized by the parameters;	for <code>broad</code>&lt=0 a minimal tree  
	 */
	public static TypedNode newTestTree(String rootType,int broad,double shrinkBy){
		return new TestTrees(rootType,broad,shrinkBy).tree;
	}
	public static String treeInfo(DataNode tree,boolean chars){
		return ""+Util.sf(!chars?Nodes.descendants(tree).length:Nodes.treeString(tree).length());
	}
	public static boolean hasMixedContents(TypedNode node){
		if(node.children().length==0)return false;
		for(Object v:node.values())if(!isKeyPair(v.toString()))return true;
		return false;
	}
}
/*/*
This file forms part of Version 0.8.79 of Facets, a framework demonstrating 
the Superficial approach to GUI-driven interactive applications.  
http://superficial.sourceforge.net 				
Copyright (C) 2011  David M Wright 
This library is free software; you can redistribute it and/or modify it under 
the terms of the GNU Lesser General Public License as published by the 
Free Software Foundation; either release 3 of the License, or (at your 
option) any later release.
This library is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR 
A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more 
details.
You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, Inc., 
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

